--*****************************************************************************
-- DISCLAIMER OF LIABILITY
-- 
-- This text/file contains proprietary, confidential
-- information of Xilinx, Inc., is distributed under license
-- from Xilinx, Inc., and may be used, copied and/or
-- disclosed only pursuant to the terms of a valid license
-- agreement with Xilinx, Inc. Xilinx hereby grants you a 
-- license to use this text/file solely for design, simulation, 
-- implementation and creation of design files limited 
-- to Xilinx devices or technologies. Use with non-Xilinx 
-- devices or technologies is expressly prohibited and 
-- immediately terminates your license unless covered by
-- a separate agreement.
--
-- Xilinx is providing this design, code, or information 
-- "as-is" solely for use in developing programs and 
-- solutions for Xilinx devices, with no obligation on the 
-- part of Xilinx to provide support. By providing this design, 
-- code, or information as one possible implementation of 
-- this feature, application or standard, Xilinx is making no 
-- representation that this implementation is free from any 
-- claims of infringement. You are responsible for 
-- obtaining any rights you may require for your implementation. 
-- Xilinx expressly disclaims any warranty whatsoever with 
-- respect to the adequacy of the implementation, including 
-- but not limited to any warranties or representations that this
-- implementation is free from claims of infringement, implied 
-- warranties of merchantability or fitness for a particular 
-- purpose.
--
-- Xilinx products are not intended for use in life support
-- appliances, devices, or systems. Use in such applications is
-- expressly prohibited.
--
-- Any modifications that are made to the Source Code are 
-- done at the user's sole risk and will be unsupported.
--
-- Copyright (c) 2007 Xilinx, Inc. All rights reserved.
--
-- This copyright and support notice must be retained as part 
-- of this text at all times. 
--*****************************************************************************
--   ____  ____
--  /   /\/   /
-- /___/  \  /    Vendor: Xilinx
-- \   \   \/     Version: 1.1
--  \   \         Filename: backend_rom.vhd
--  /   /         Timestamp: 24 August 2007
-- /___/   /\     
-- \   \  /  \
--  \___\/\___\
--
--
--	Device:  Virtex-4 and Virtex-5
--	Purpose: This module generates the Write Address enable, Write Data 
--		enable and Read Address enable signals, along with the Write Data
--		and Write and Read addresses themselves.
--		It also provides comparison data to the data checker to be 
--		compared with the data that is read via the Read FIFO.
-- Revision History:
--   Rev 1.0 - initial release
--   Rev 1.1 - Added the RL_DQ_WIDTH parameter, generate write data for different widths
--*****************************************************************************

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;

library UNISIM;
use UNISIM.VCOMPONENTS.ALL;


ENTITY backend_rom IS
   GENERIC (
      RL_DQ_WIDTH                    : integer := 36;   -- memory interface data width
      DEV_DQ_WIDTH                   : integer := 18;   -- data width of the memory device
      NUM_OF_DEVS                    : integer := 2;  -- number of memory devices
      NUM_OF_DKS                     : integer := 4;
      DEV_AD_WIDTH                   : integer := 20;   -- address width of the memory device
      DEV_BA_WIDTH                   : integer := 3;    -- bank address width of the memory device
      APP_AD_WIDTH                   : integer := 26; 
      DEV_BYTE_WIDTH                 : integer := 9;    -- the width of a "byte" on the memory device (x8 or x9)
      RL_MRS_BURST_LENGTH            : std_logic_vector(1 DOWNTO 0) := "00"     -- "00": BL2;  "01": BL4;  "10": BL8
   );
   PORT (
      clk                     : IN  std_logic;
      rst                     : IN  std_logic;
      rst90                   : IN  std_logic;
      clk90                   : IN  std_logic;
      bkend_data_en           : IN  std_logic;
      bkend_wraddr_en         : IN  std_logic;
      bkend_rdaddr_en         : IN  std_logic;
      apRdfRdEn               : IN  std_logic;
      burst_length            : IN  std_logic_vector(1 DOWNTO 0);
      app_Waf_addr            : OUT std_logic_vector(APP_AD_WIDTH-1 DOWNTO 0);
      app_Waf_WrEn            : OUT std_logic;
      app_Wdf_data            : OUT std_logic_vector((2*RL_DQ_WIDTH)-1 DOWNTO 0);
      app_compare_data        : OUT std_logic_vector((2*RL_DQ_WIDTH)-1 DOWNTO 0);
      app_Wdf_WrEn            : OUT std_logic;
      app_Raf_addr            : OUT std_logic_vector(APP_AD_WIDTH-1 DOWNTO 0);
      app_Raf_WrEn            : OUT std_logic;
      cs_io                   : INOUT std_logic_vector(1023 DOWNTO 0)
   );
END ENTITY backend_rom;



ARCHITECTURE rtl OF backend_rom IS


--   -- Data generation state machine (state encodings)
--   CONSTANT IDLE_FIRST_DATA         : std_logic_vector(1 DOWNTO 0) := "00";
--   CONSTANT SECOND_DATA             : std_logic_vector(1 DOWNTO 0) := "01";
--   CONSTANT THIRD_DATA              : std_logic_vector(1 DOWNTO 0) := "10";
--   CONSTANT FOURTH_DATA             : std_logic_vector(1 DOWNTO 0) := "11";

   -- Data generation state machine (state encodings)
   type dataGenStates is ( IDLE_FIRST_DATA, SECOND_DATA, THIRD_DATA, FOURTH_DATA );
   SIGNAL state    : dataGenStates;
   SIGNAL state_90 : dataGenStates;


   SIGNAL wr_rd_addr               : std_logic_vector(8 DOWNTO 0);  -- 9-bit value
   SIGNAL wr_rd_addr_en            : std_logic;
   SIGNAL addr_count               : std_logic_vector(4 DOWNTO 0);
--   SIGNAL state                    : std_logic_vector(1 DOWNTO 0);
--   SIGNAL state_90                 : std_logic_vector(1 DOWNTO 0);
   SIGNAL data_pattern             : std_logic_vector(8 DOWNTO 0);
   SIGNAL data_pattern_90          : std_logic_vector(8 DOWNTO 0);
   SIGNAL app_Waf_addr_1           : std_logic_vector(APP_AD_WIDTH-1 DOWNTO 0);
   SIGNAL app_Raf_addr_1           : std_logic_vector(APP_AD_WIDTH-1 DOWNTO 0);
   SIGNAL app_Wdf_WrEn_a1          : std_logic;
   SIGNAL rising_edge_data         : std_logic_vector(8 DOWNTO 0);
   SIGNAL falling_edge_data        : std_logic_vector(8 DOWNTO 0);
   SIGNAL rising_edge_data_90      : std_logic_vector(8 DOWNTO 0);
   SIGNAL falling_edge_data_90     : std_logic_vector(8 DOWNTO 0);

   SIGNAL unused_data_in           : std_logic_vector(31 DOWNTO 0);
   SIGNAL unused_data_in_p         : std_logic_vector(3 DOWNTO 0);

   SIGNAL gnd                      : std_logic;
   SIGNAL vcc                      : std_logic;

   SIGNAL addr_out                 : std_logic_vector(31 DOWNTO 0);
   SIGNAL app_Wdf_data_a1          : std_logic_vector((2*RL_DQ_WIDTH)-1 DOWNTO 0);

   -- create pulse to re-init BRAM dout after reset
   SIGNAL capture_rst              : std_logic;
   SIGNAL dly_capture_rst          : std_logic;
   SIGNAL dly2_capture_rst         : std_logic;
   SIGNAL init_BRAM_din            : std_logic;
   SIGNAL init_BRAM_dout           : std_logic;
   SIGNAL bram_sinit               : std_logic;

   -- Chipscope assignemnet
   --cs_io(190 DOWNTO 173) <= "00000000000000000000" & addr_count & wr_rd_addr_en;

   -- ADDED, RCHIU
   SIGNAL ramb36_addra             : std_logic_vector(15 DOWNTO 0);
   SIGNAL fixed_bits               : std_logic_vector(15 DOWNTO 0);


BEGIN
   cs_io <= (OTHERS => 'Z');
   unused_data_in    <= (others => '0');   -- 32 bits
   unused_data_in_p  <= (others => '0');   -- 4 bits

   gnd <= '0';
   vcc <= '1';


   -- -----------------------------------------------------------------------------
   -- Address generation for Write and Read Address FIFOs
   -- -----------------------------------------------------------------------------
   -- apAddr(31:26) [ 6 bits] - not used...
   -- apAddr(25)    [ 1 bit ] - user refresh
   -- apAddr(24)    [ 1 bit ] - R/~W  ( 0: write command;  1: read command )
   -- apAddr(23:3)  [21 bits] - memory address  A(21-1:0)
   -- apAddr(2:0)   [ 3 bits] - memory bank address  BA(2:0)
   -- -----------------------------------------------------------------------------
   --
   -- ROM with address patterns
   --
   -- The address patterns below in INIT strings support ((8 Writes & 8 Reads) + another (8 Writes & 8 Reads)).
   -- New INIT values, hence new Read/Write address locations, can be added if desired.
   --
   -- INIT_00 = 256'h0100000000000007000000060000000500000004000000030000000200000001,
   -- INIT_01 = 256'h0000000801000007010000060100000501000004010000030100000201000001,
   -- INIT_02 = 256'h010000080000000F0000000E0000000D0000000C0000000B0000000A00000009,
   -- INIT_03 = 256'h000000000100000F0100000E0100000D0100000C0100000B0100000A01000009;	   
   --
   -- --------------------------------------------------------------------
   -- ** Details of INIT_00 **
   -- --------------------------------------------------------------------
   --        7654 3210  -- Note: bits [31:26] are unused --
   --        |||| ||||  bit 25 No user refresh, bit 24 READ Command,   bits [23:3] 21-bits A=0, bits [2:0] bank address BA=0
   -- [0x07] 0100 0000  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0000
   --                   bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=0, bits [2:0] bank address BA=7
   -- [0x06] 0000 0007  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0111
   --                   bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=0, bits [2:0] bank address BA=6
   -- [0x05] 0000 0006  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0110
   --                   bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=0, bits [2:0] bank address BA=5
   -- [0x04] 0000 0005  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0101
   --                   bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=0, bits [2:0] bank address BA=4
   -- [0x03] 0000 0004  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0100
   --                   bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=0, bits [2:0] bank address BA=3
   -- [0x02] 0000 0003  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0011
   --                   bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=0, bits [2:0] bank address BA=2
   -- [0x01] 0000 0002  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0010
   --                   bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=0, bits [2:0] bank address BA=1
   -- [0x00] 0000 0001  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0001
   -- 
   -- --------------------------------------------------------------------
   -- ** Details of INIT_01 **
   -- --------------------------------------------------------------------
   --        7654 3210  -- Note: bits [31:26] are unused --
   --        |||| ||||  bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=1, bits [2:0] bank address BA=0
   -- [0x0F] 0000 0008  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1000
   --                   bit 25 No user refresh, bit 24 READ  Command,  bits [23:3] 21-bits A=0, bits [2:0] bank address BA=7
   -- [0x0E] 0100 0007  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0111
   --                   bit 25 No user refresh, bit 24 READ Command,   bits [23:3] 21-bits A=0, bits [2:0] bank address BA=6
   -- [0x0D] 0100 0006  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0110
   --                   bit 25 No user refresh, bit 24 READ Command,   bits [23:3] 21-bits A=0, bits [2:0] bank address BA=5
   -- [0x0C] 0100 0005  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0101
   --                   bit 25 No user refresh, bit 24 READ Command,   bits [23:3] 21-bits A=0, bits [2:0] bank address BA=4
   -- [0x0B] 0100 0004  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0100
   --                   bit 25 No user refresh, bit 24 READ Command,   bits [23:3] 21-bits A=0, bits [2:0] bank address BA=3
   -- [0x0A] 0100 0003  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0011
   --                   bit 25 No user refresh, bit 24 READ Command,   bits [23:3] 21-bits A=0, bits [2:0] bank address BA=2
   -- [0x09] 0100 0002  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0010
   --                   bit 25 No user refresh, bit 24 READ Command,   bits [23:3] 21-bits A=0, bits [2:0] bank address BA=1
   -- [0x08] 0100 0001  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0001
   -- 
   -- --------------------------------------------------------------------
   -- ** Details of INIT_02 **
   -- --------------------------------------------------------------------
   --        7654 3210  -- Note: bits [31:26] are unused --
   --        |||| ||||  bit 25 No user refresh, bit 24 READ Command,   bits [23:3] 21-bits A=1, bits [2:0] bank address BA=0
   -- [0x17] 0100 0008  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1000
   --                   bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=1, bits [2:0] bank address BA=7
   -- [0x16] 0000 000F  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1111
   --                   bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=1, bits [2:0] bank address BA=6
   -- [0x15] 0000 000E  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1110
   --                   bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=1, bits [2:0] bank address BA=5
   -- [0x14] 0000 000D  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1101
   --                   bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=1, bits [2:0] bank address BA=4
   -- [0x13] 0000 000C  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1100
   --                   bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=1, bits [2:0] bank address BA=3
   -- [0x12] 0000 000B  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1011
   --                   bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=1, bits [2:0] bank address BA=2
   -- [0x11] 0000 000A  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1010
   --                   bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=1, bits [2:0] bank address BA=1
   -- [0x10] 0000 0009  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1001
   -- 
   -- --------------------------------------------------------------------
   -- ** Details of INIT_03 **
   -- --------------------------------------------------------------------
   --        7654 3210  -- Note: bits [31:26] are unused --
   --        |||| ||||  bit 25 No user refresh, bit 24 WRITE Command,  bits [23:3] 21-bits A=0, bits [2:0] bank address BA=0
   -- [0x1F] 0000 0000  [31:28] 0000, [27:24] 0000, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 0000
   --                   bit 25 No user refresh, bit 24 READ  Command,  bits [23:3] 21-bits A=1, bits [2:0] bank address BA=7
   -- [0x1E] 0100 000F  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1111
   --                   bit 25 No user refresh, bit 24 READ Command,   bits [23:3] 21-bits A=1, bits [2:0] bank address BA=6
   -- [0x1D] 0100 000E  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1110
   --                   bit 25 No user refresh, bit 24 READ Command,   bits [23:3] 21-bits A=1, bits [2:0] bank address BA=5
   -- [0x1C] 0100 000D  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1101
   --                   bit 25 No user refresh, bit 24 READ Command,   bits [23:3] 21-bits A=1, bits [2:0] bank address BA=4
   -- [0x1B] 0100 000C  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1100
   --                   bit 25 No user refresh, bit 24 READ Command,   bits [23:3] 21-bits A=1, bits [2:0] bank address BA=3
   -- [0x1A] 0100 000B  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1011
   --                   bit 25 No user refresh, bit 24 READ Command,   bits [23:3] 21-bits A=1, bits [2:0] bank address BA=2
   -- [0x19] 0100 000A  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1010
   --                   bit 25 No user refresh, bit 24 READ Command,   bits [23:3] 21-bits A=1, bits [2:0] bank address BA=1
   -- [0x18] 0100 0009  [31:28] 0000, [27:24] 0001, [23:20] 0000, [19:16] 0000, [15:12] 0000, [11:8] 0000, [7:4] 0000, [3:0] 1001
   -- 


   -- -----------------------------------------------------------------------------
   -- New RAMB36 support, for Virtex-4 and Virtex-5
   -- -----------------------------------------------------------------------------
    wr_rd_addr_lookup : RAMB36 
      GENERIC MAP (
--         DOA_REG      => 1,   -- Optional output registers on the A port (0 or 1)
--         DOB_REG      => 0,   -- Optional output registers on the B port (0 or 1)
         READ_WIDTH_A => 36,
         READ_WIDTH_B => 36,
--         INIT_A       => "000000000000000000000000000000000000",  -- (36-bit) Initial values on A output port
--         INIT_B       => "000000000000000000000000000000000000",  -- (36-bit) Initial values on B output port
--         SRVAL_A      => "000000000000000000000000000000000000",  -- (36-bit) Set/Reset value for A port output
--         SRVAL_B      => "000000000000000000000000000000000000",  -- (36-bit) Set/Reset value for B port output
         INIT_00      => X"0100000000000007000000060000000500000004000000030000000200000001",
         INIT_01      => X"0000000801000007010000060100000501000004010000030100000201000001",
         INIT_02      => X"010000080000000F0000000E0000000D0000000C0000000B0000000A00000009",
         INIT_03      => X"000000000100000F0100000E0100000D0100000C0100000B0100000A01000009"
      )
      PORT MAP (
         CASCADEOUTLATA  => open,
         CASCADEOUTLATB  => open,
         CASCADEOUTREGA  => open,
         CASCADEOUTREGB  => open,
         DOA             => addr_out(31 DOWNTO 0),
         DOB             => open,
         DOPA            => open,
         DOPB            => open,
         ADDRA           => ramb36_addra,
         ADDRB           => "0000000000000000",  -- 16-bit to zero
         CASCADEINLATA   => gnd,   -- (unused)
         CASCADEINLATB   => gnd,   -- (unused)
         CASCADEINREGA   => gnd,   -- (unused)
         CASCADEINREGB   => gnd,   -- (unused)
         CLKA            => clk,
         CLKB            => clk,
         DIA             => unused_data_in,      -- 32-bit to zero
         DIB             => unused_data_in,      -- 32-bit to zero
         DIPA            => unused_data_in_p,    --  4-bit to zero
         DIPB            => unused_data_in_p,    --  4-bit to zero
         -- --------------------------------------------------------------------
         -- Warning when ENA and/or ENB are always enabled :
         --    Must make sure to get synchronous address even when WEA/WEB
         --    are disabled, else potential corruption could happen.
         -- --------------------------------------------------------------------
         -- See Answer Record #21870 for details.
         -- --------------------------------------------------------------------
--         ENA             => vcc,  -- always enable
--         ENA             => app_Waf_WrEn OR app_Raf_WrEn,
         ENA             => wr_rd_addr_en,
         ENB             => vcc,  -- always enable
         -- --------------------------------------------------------------------
         REGCEA          => gnd,  -- (unused)
         REGCEB          => gnd,  -- (unused)
         SSRA            => bram_sinit,
         SSRB            => bram_sinit,
         WEA             => "0000",              -- 4-bit to zero
         WEB             => "0000"               -- 4-bit to zero
      );


   -- must pulse the BRAM SINIT upon reset to get correct DOUT value
   TEST_fd_bram_sinit : FD 
   PORT MAP (
      C => clk,
      D => init_BRAM_dout,
      Q => bram_sinit
   );


   -- enable for ROM with address patterns
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (rst = '1') THEN
            wr_rd_addr_en   <= '0';
            app_Wdf_WrEn_a1 <= '0';
         ELSE
            app_Wdf_WrEn_a1 <= bkend_data_en;
            wr_rd_addr_en   <= bkend_wraddr_en OR bkend_rdaddr_en OR init_BRAM_dout;   -- use delayed reset pulse to get BRAM's dout re-initialized
         END IF;
      END IF;
   END PROCESS;


   -- capture the reset
   fd_capture_rst : FD
   PORT MAP (
      C => clk,
      D => rst,
      Q => capture_rst
   );


   -- delay the captured reset
   delayed_capture_rst : SRL16
   PORT MAP (
      Q   => dly_capture_rst,
      A0  => gnd,
      A1  => vcc,
      A2  => vcc,
      A3  => vcc,
      CLK => clk,
      D   => capture_rst
   );


   -- one more delay
   fd_delayed_capture_rst : FD
   PORT MAP (
      C => clk,
      D => dly_capture_rst,
      Q => dly2_capture_rst
   );


   -- generate pulse on falling edge
   init_BRAM_din <= NOT dly_capture_rst AND dly2_capture_rst;

   fd_init_BRAM_dout : FD
   PORT MAP (
      C => clk,
      D => init_BRAM_din,   -- capture falling edge
      Q => init_BRAM_dout
   );


   -- delay signals
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (rst = '1') THEN
            app_Wdf_data <= (others => '0');
            app_Wdf_WrEn <= '0';
         ELSE
            app_Wdf_data <= app_Wdf_data_a1;
            app_Wdf_WrEn <= app_Wdf_WrEn_a1;
         END IF;
      END IF;
   END PROCESS;


   -- output read and write enables and addresses
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (rst = '1') THEN
            app_Waf_addr_1 <= (others => '0');
            app_Waf_WrEn   <= '0';
            app_Raf_WrEn   <= '0';
            app_Raf_addr_1 <= (others => '0');
         ELSE
            app_Waf_addr_1 <= addr_out(APP_AD_WIDTH-1 DOWNTO 0);
            app_Waf_WrEn   <= bkend_wraddr_en;
            app_Raf_WrEn   <= bkend_rdaddr_en;
            app_Raf_addr_1 <= addr_out(APP_AD_WIDTH-1 DOWNTO 0);
         END IF;
      END IF;
   END PROCESS;


   app_Waf_addr <= addr_out(APP_AD_WIDTH-1 DOWNTO 0) WHEN (burst_length = "00") ELSE app_Waf_addr_1;

   app_Raf_addr <= addr_out(APP_AD_WIDTH-1 DOWNTO 0) WHEN (burst_length = "00") ELSE app_Raf_addr_1;


   -- address locations in ROM
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (rst = '1') THEN
            addr_count <= (others => '0');
            wr_rd_addr <= (others => '0');
         ELSE
            IF ( (bkend_wraddr_en OR bkend_rdaddr_en) = '1' ) THEN
               addr_count <= STD_LOGIC_VECTOR( UNSIGNED(addr_count) + 1 );   -- increment      
      -- ADDED, RCHIU, 03/15/06 - force all bits of wr_rd_addr to be "nonconstant"
               -- wr_rd_addr <= "0000" & addr_count;
               wr_rd_addr <= fixed_bits(13 DOWNTO 10) & addr_count;
            ELSE
               addr_count <= addr_count;
               wr_rd_addr <= wr_rd_addr;
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   --
   -- RAMB36 address width is constant with 16 bits
   --
   -- -----------------------------------------------------------------------------
   -- <<< 9-bit address value >>>
   -- -----------------------------------------------------------------------------
   -- ramb36_addra  |- [15:14] -|--- [13:5] ---|--------- [4:0] ---------|
   --               |  "unused" | "wr_rd_addr" | "32-bit data locations" |
   --
   -- -> ramb36_addra [15:14] => unused (2 bits)
   -- -> ramb36_addra [13:5]  => wr_rd_addr, 512 address locations needs 9 bits
   -- -> ramb36_addra [4:0]   => access individual data bit locations, 
   --                            5 bits for 32 bits data?
   --
   -- -----------------------------------------------------------------------------
   -- <<< 10-bit address value >>>
   -- -----------------------------------------------------------------------------
   -- ramb36_addra  |-- [15] --|--- [14:5] ---|--------- [4:0] ---------|
   --               | "unused" | "wr_rd_addr" | "32-bit data locations" |
   --
   -- -> ramb36_addra [15]    => unused (1 bit)
   -- -> ramb36_addra [14:5]  => wr_rd_addr, 1024 address locations needs 10 bits
   -- -> ramb36_addra [4:0]   => access individual data bit locations, 
   --                            5 bits for 32 bits data?
   --
   -- -----------------------------------------------------------------------------
   -- ADDED, RCHIU, get around RAMB36 initialization bug
   -- RAMB36 doesn't initialize when upper bits of address are fixed to 0
   ramb36_addra <= fixed_bits(15 DOWNTO 14) & wr_rd_addr & "00000";

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (rst = '1') THEN
            -- just a hack to try and get around synthesis optimizing this out
            fixed_bits <= (others => '1');
         ELSE
            fixed_bits <= (others => '0');
         END IF;
      END IF;
   END PROCESS;


   -- DATA generation for write data fifos
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (rst = '1') THEN
            data_pattern <= (others => '0');    -- 0x000 (9 bits)
            state        <= IDLE_FIRST_DATA;
         ELSE
            CASE state IS
               -- --------------------------------------------------------------
               WHEN IDLE_FIRST_DATA =>
               -- --------------------------------------------------------------
                  IF (bkend_data_en = '1') THEN
                     data_pattern <= "111101110";    -- 0x1EE (9 bits)
                     state        <= SECOND_DATA;
                  ELSE
                     state        <= IDLE_FIRST_DATA;
                  END IF;
               -- --------------------------------------------------------------
               WHEN SECOND_DATA =>
               -- --------------------------------------------------------------
                  IF (bkend_data_en = '1') THEN
                     data_pattern <= "110101010";    -- 0x1AA (9 bits)
                     state        <= IDLE_FIRST_DATA;
                  END IF;
               -- --------------------------------------------------------------
               WHEN OTHERS =>
               -- --------------------------------------------------------------
                  NULL;
            END CASE;
         END IF;
      END IF;
   END PROCESS;


   -- data generation for read data compare
   -- data read from memory is in system clock domain
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (rst = '1') THEN
            data_pattern_90 <= (others => '0');    -- 0x000 (9 bits)
            state_90        <= IDLE_FIRST_DATA;
         ELSE
            CASE state_90 IS
               -- --------------------------------------------------------------
               WHEN IDLE_FIRST_DATA =>
               -- --------------------------------------------------------------
                  IF ( apRdfRdEn = '1' ) THEN
                     data_pattern_90 <= "111101110";    -- 0x1EE (9 bits)
                     state_90        <= SECOND_DATA;
                  ELSE
                     state_90        <= IDLE_FIRST_DATA;
                  END IF;
               -- --------------------------------------------------------------
               WHEN SECOND_DATA =>
               -- --------------------------------------------------------------
                  IF ( apRdfRdEn = '1' ) THEN
                     data_pattern_90 <= "110101010";    -- 0x1AA (9 bits)
                     state_90        <= IDLE_FIRST_DATA;
                  END IF;
               -- --------------------------------------------------------------
               WHEN OTHERS =>
               -- --------------------------------------------------------------
                  NULL;
            END CASE;
         END IF;
      END IF;
   END PROCESS;


   rising_edge_data     <= data_pattern;
   falling_edge_data    <= NOT data_pattern;

   rising_edge_data_90  <= data_pattern_90;
   falling_edge_data_90 <= NOT data_pattern_90;

   back_rom_18 : IF ( RL_DQ_WIDTH = 18) GENERATE
   BEGIN
   -- data to write data fifos during write
   -- 2x same 18-bit data
   app_Wdf_data_a1 <= rising_edge_data  & rising_edge_data & 
                      falling_edge_data & falling_edge_data;

   -- data to compare circuit during read
   -- 2x same 18-bit data
   app_compare_data <= rising_edge_data_90 & rising_edge_data_90 & 
                       falling_edge_data_90 & falling_edge_data_90;
                       
   END GENERATE back_rom_18;
   
   back_rom_36 : IF ( RL_DQ_WIDTH = 36) GENERATE
   BEGIN
   -- data to write data fifos during write
   -- 4x same 18-bit data
   app_Wdf_data_a1 <= rising_edge_data  & rising_edge_data  & rising_edge_data  & rising_edge_data & 
                      falling_edge_data & falling_edge_data & falling_edge_data & falling_edge_data;

   -- data to compare circuit during read
   -- 4x same 18-bit data
   app_compare_data <= rising_edge_data_90  & rising_edge_data_90  & rising_edge_data_90  & rising_edge_data_90 & 
                       falling_edge_data_90 & falling_edge_data_90 & falling_edge_data_90 & falling_edge_data_90;

   END GENERATE back_rom_36;
   
   back_rom_72 : IF ( RL_DQ_WIDTH = 72) GENERATE
   BEGIN
   -- data to write data fifos during write
   -- 8x same 18-bit data
   app_Wdf_data_a1 <= rising_edge_data  & rising_edge_data  & rising_edge_data  & rising_edge_data & 
                      rising_edge_data  & rising_edge_data  & rising_edge_data  & rising_edge_data & 
                      falling_edge_data & falling_edge_data & falling_edge_data & falling_edge_data &
                      falling_edge_data & falling_edge_data & falling_edge_data & falling_edge_data;

   -- data to compare circuit during read
   -- 8x same 18-bit data
   app_compare_data <= rising_edge_data_90  & rising_edge_data_90  & rising_edge_data_90  & rising_edge_data_90 &
                       rising_edge_data_90  & rising_edge_data_90  & rising_edge_data_90  & rising_edge_data_90 & 
                       falling_edge_data_90 & falling_edge_data_90 & falling_edge_data_90 & falling_edge_data_90 &
                       falling_edge_data_90 & falling_edge_data_90 & falling_edge_data_90 & falling_edge_data_90;
                       
   END GENERATE back_rom_72;
   
   

END ARCHITECTURE rtl;
