--*****************************************************************************
-- DISCLAIMER OF LIABILITY
-- 
-- This text/file contains proprietary, confidential
-- information of Xilinx, Inc., is distributed under license
-- from Xilinx, Inc., and may be used, copied and/or
-- disclosed only pursuant to the terms of a valid license
-- agreement with Xilinx, Inc. Xilinx hereby grants you a 
-- license to use this text/file solely for design, simulation, 
-- implementation and creation of design files limited 
-- to Xilinx devices or technologies. Use with non-Xilinx 
-- devices or technologies is expressly prohibited and 
-- immediately terminates your license unless covered by
-- a separate agreement.
--
-- Xilinx is providing this design, code, or information 
-- "as-is" solely for use in developing programs and 
-- solutions for Xilinx devices, with no obligation on the 
-- part of Xilinx to provide support. By providing this design, 
-- code, or information as one possible implementation of 
-- this feature, application or standard, Xilinx is making no 
-- representation that this implementation is free from any 
-- claims of infringement. You are responsible for 
-- obtaining any rights you may require for your implementation. 
-- Xilinx expressly disclaims any warranty whatsoever with 
-- respect to the adequacy of the implementation, including 
-- but not limited to any warranties or representations that this
-- implementation is free from claims of infringement, implied 
-- warranties of merchantability or fitness for a particular 
-- purpose.
--
-- Xilinx products are not intended for use in life support
-- appliances, devices, or systems. Use in such applications is
-- expressly prohibited.
--
-- Any modifications that are made to the Source Code are 
-- done at the user's sole risk and will be unsupported.
--
-- Copyright (c) 2007 Xilinx, Inc. All rights reserved.
--
-- This copyright and support notice must be retained as part 
-- of this text at all times. 
--*****************************************************************************
--   ____  ____
--  /   /\/   /
-- /___/  \  /    Vendor: Xilinx
-- \   \   \/     Version: 1.1
--  \   \         Filename: rld_ctl.vhd
--  /   /         Timestamp: 24 August 2007
-- /___/   /\     
-- \   \  /  \
--  \___\/\___\
--
--
--	Device:  Virtex-4 and Virtex-5
--	Purpose: This module is the memory controller;
--		* Main control is implemented in this module.
--		* It initiates read, write and refresh commands to rld memory.
-- Revision History:
--   Rev 1.0 - initial release
--   Rev 1.1 - changed the clock frequency parameter to a clock period parameter,
--             added the X_CORE synthesis attribute
--
--		This module processes commands from the command FIFO and sends
--		the commands to the memory interface.  It also handles 
--		auto-refresh and bank conflict.
--
--  ** Note *******************************************************************
--  1. This module has been extensively tested with random stimulus for BL=2
--     and CONF=1.  Limited tests were run for other configurations and there
--     may be uncovered problems with these configurations.  More tests will
--     be done for all supported configurations in the next phase.
--
--  2. User refresh command is not supported in this version.  Please use AREF 
--     feature provided by the controller instead.
--  ***************************************************************************
--
--
--  Port Definitions:
--     clk              (input)  clock input
--     rstHard          (input)  hard reset, active high
--     rstConfig        (input)  configuration reset
--
--     cmdFifoEmpty     (input)  command FIFO empty flag
--     cmdFifoOut[25:0] (input)  command FIFO data output
--          - [25:24]: command 00: write;  01: read
--          - [23:3] : memory address (A)
--          - [2:0]  : memory bank address (BA)
--     rstCmd           (input)  command during configuration reset
--     rstAd            (input)  memory address during configuration reset
--     rstBa            (input)  memory bank during configuration reset
--
--     confRL           (input)  memory read latency
--     confWL           (input)  memory write latency
--     confRcCnt0       (input)
--     confRcCnt1       (input) 
--          - (confRcCnt1, confRcCnt0} is the interval (in clock cycles) 
--            that AREF should be issued
--     confTrc          (input)  memory tRC setting
--     confRcMargin     (input)  confRcMargin defines how many cycles before 
--          - {confRcCnt1, confRcCnt0} expires that a AREF request should be sent to the FSM
--     confTrdWr        (input)  unused...
--     confTwrRd        (input)  unused...
--     confBL           (input)  memory burst length
--     confAutoRefrEn   (input)  enables AREF feature of the controller
--
--     cmdFifoRdEn     (output)  command FIFO read enable
--     ctlIoWdEn       (output)  IO output eanble
--     ctlCmdOut       (output)  memory command 
--     ctlAddrOut      (output)  memory address
--     ctlWdfRdEn      (output)  write data FIFO read enable
--     ctlBankOut      (output)  memory bank
--     ctlOkEdge       (output)  indicates OK to change data valid edge for per-bit deskew
--     rldWriteDone    (output)  indicates one write command was executed
--
--*****************************************************************************

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;


ENTITY rld_ctl IS
   GENERIC (
      RL_DQ_WIDTH             : integer := 36;     -- memory interface data width
      DEV_DQ_WIDTH            : integer := 18;     -- data width of the memory device
      NUM_OF_DEVS             : integer := 2;      -- number of memory devices
      NUM_OF_DKS              : integer := 4; 
      DEV_AD_WIDTH            : integer := 20;     -- address width of the memory device
      DEV_BA_WIDTH            : integer := 3;      -- bank address width of the memory device
      APP_AD_WIDTH            : integer := 26;
      DUPLICATE_CONTROLS      : boolean := FALSE;  -- Duplicate the ports for A, BA, WE# and REF#
      -- 
      RL_CK_PERIOD            : integer := 3003;   -- CK clock period of the RLDRAM in ps
      -- MRS (Mode Register Set command) parameters   
      --    please check Micron RLDRAM-II datasheet for definitions of these parameters
      RL_MRS_CONF             : std_logic_vector(2 DOWNTO 0) := "010";   -- "001": mode1;  "010": mode2;  "011": mode3
      RL_MRS_BURST_LENGTH     : std_logic_vector(1 DOWNTO 0) := "01";    -- "00": BL2;  "01": BL4;  "10": BL8
      RL_MRS_ADDR_MUX         : std_logic := '0';    -- '0': non-muxed addr;  '1': muxed addr
      RL_MRS_DLL_RESET        : std_logic := '0';    -- 
      RL_MRS_IMPEDANCE_MATCH  : std_logic := '1';    -- '0': internal 50ohms output buffer impedance, '1': external 
      RL_MRS_ODT              : std_logic := '0';    -- '0': disable term;  '1': enable term
      -- specific to FPGA/memory devices and capture method
      RL_IO_TYPE              : std_logic_vector(1 DOWNTO 0) := "00";  -- CIO="00"  SIO_CIO_HYBRID="01"  SIO="10"
      DEVICE_ARCH             : std_logic_vector(1 DOWNTO 0) := "01";  -- Virtex4="00"  Virtex5="01"
      CAPTURE_METHOD          : std_logic_vector(1 DOWNTO 0) := "01"   -- Direct Clocking="00"  SerDes="01"
   );
   PORT (
      clk                     : IN std_logic;
      rstHard                 : IN std_logic;
      rstConfig               : IN std_logic;
      cmdFifoEmpty            : IN std_logic;
      cmdFifoOut              : IN std_logic_vector(25 DOWNTO 0);
      -- Reset configuration signals
      rstCmd                  : IN std_logic_vector(3 DOWNTO 0);
      rstAd                   : IN std_logic_vector(DEV_AD_WIDTH-1 DOWNTO 0);
      rstBa                   : IN std_logic_vector(DEV_BA_WIDTH-1 DOWNTO 0);
      -- Configuration Register values
      confRL                  : IN std_logic_vector(3 DOWNTO 0);
      confWL                  : IN std_logic_vector(3 DOWNTO 0);
      confRcCnt0              : IN std_logic_vector(7 DOWNTO 0);
      confRcCnt1              : IN std_logic_vector(7 DOWNTO 0);
      confRcMargin            : IN std_logic_vector(7 DOWNTO 0);
      confTrc                 : IN std_logic_vector(3 DOWNTO 0);
      confTrdWr               : IN std_logic_vector(2 DOWNTO 0);
      confTwrRd               : IN std_logic_vector(2 DOWNTO 0);
      confBL                  : IN std_logic_vector(3 DOWNTO 0);
      confAutoRefrEn          : IN std_logic;
      cmdFifoRdEn             : OUT std_logic;
      -- write data enable for pads
      ctlIoWdEn               : OUT std_logic;
      -- commands and address, bank outputs
      ctlCmdOut               : OUT std_logic_vector(3 DOWNTO 0);
      ctlAddrOut              : OUT std_logic_vector(DEV_AD_WIDTH-1 DOWNTO 0);
      -- write data fifo read
      ctlWdfRdEn              : OUT std_logic;
      ctlBankOut              : OUT std_logic_vector(DEV_BA_WIDTH-1 DOWNTO 0);
      ctlOkEdge               : OUT std_logic;
      rldWriteDone            : OUT std_logic;
      -- chipscope debug ports
      cs_io                   : INOUT std_logic_vector(1023 DOWNTO 0)
   );
   ATTRIBUTE X_CORE_INFO : STRING;
   ATTRIBUTE X_CORE_INFO of rld_ctl : ENTITY IS "rld_ctl, V5_rldram2_cont_vhdl";   
END ENTITY rld_ctl;



ARCHITECTURE rtl OF rld_ctl IS

   function OR_REDUCE(ARG: in STD_LOGIC_VECTOR) return std_logic is
      variable result: STD_LOGIC := '0';
   begin
      for i in ARG'range loop
         result := result OR ARG(i);
      end loop;
      return result;
   end function OR_REDUCE;

   function AND_REDUCE(ARG: in STD_LOGIC_VECTOR) return std_logic is
      variable result: STD_LOGIC := '1';
   begin
      for i in ARG'range loop
         result := result AND ARG(i);
      end loop;
      return result;
   end function AND_REDUCE;


   -- Controller state machine (state encodings)
--   CONSTANT CTL_IDLE           : std_logic_vector(2 DOWNTO 0) := "000";
--   CONSTANT CTL_LOAD_CMD1      : std_logic_vector(2 DOWNTO 0) := "001";
--   CONSTANT CTL_LOAD_CMD2      : std_logic_vector(2 DOWNTO 0) := "010";
--   CONSTANT CTL_PROC_CMD       : std_logic_vector(2 DOWNTO 0) := "011";
--   CONSTANT CTL_PROC_LAST_CMD  : std_logic_vector(2 DOWNTO 0) := "100";
--   CONSTANT CTL_PROC_REFR      : std_logic_vector(2 DOWNTO 0) := "101";
   type CtlStates is ( CTL_IDLE, CTL_LOAD_CMD1, CTL_LOAD_CMD2, CTL_PROC_CMD, CTL_PROC_LAST_CMD, CTL_PROC_REFR );
   SIGNAL ctl_state  : CtlStates;

   -- CMD signals are formed by concatenating CS_N, AS_N, WE_N, REF_N, respectively
   --                                                   CS_N  AS_N  WE_N  REF_N
   CONSTANT CMD_DESEL  : std_logic_vector(3 DOWNTO 0) := '1' & '1' & '1' & '1';
   CONSTANT CMD_RD     : std_logic_vector(3 DOWNTO 0) := '0' & '0' & '1' & '1';
   CONSTANT CMD_WR     : std_logic_vector(3 DOWNTO 0) := '0' & '0' & '0' & '1';
   CONSTANT CMD_NOP    : std_logic_vector(3 DOWNTO 0) := '1' & '1' & '1' & '1';
   CONSTANT CMD_REF    : std_logic_vector(3 DOWNTO 0) := '0' & '1' & '1' & '0';


   SIGNAL load_cmd1                : std_logic;
   SIGNAL load_cmd2                : std_logic;

   SIGNAL rst                      : std_logic;
   SIGNAL rd_grant_sreg            : std_logic_vector(14 DOWNTO 0);
--   SIGNAL refUser                  : std_logic;   -- register for user refresh, unsupported for now
   SIGNAL refr_cnt                 : std_logic_vector(11 DOWNTO 0);
   SIGNAL refr_req                 : std_logic;
   SIGNAL refr_req_clr             : std_logic;
   SIGNAL refr_trc_ok              : std_logic;
   SIGNAL refr_req_clr_sreg        : std_logic_vector(7 DOWNTO 0);

   SIGNAL refr_bank                : std_logic_vector(2 DOWNTO 0);
   SIGNAL refr_done                : std_logic;

   SIGNAL cmd_addr_r1              : std_logic_vector(DEV_AD_WIDTH-1 DOWNTO 0);
   SIGNAL cmd_cmd_r1               : std_logic_vector(1 DOWNTO 0);
   SIGNAL cmd_bank_r1              : std_logic_vector(DEV_BA_WIDTH-1 DOWNTO 0);

   SIGNAL cmd_addr_r2              : std_logic_vector(DEV_AD_WIDTH-1 DOWNTO 0);
   SIGNAL cmd_cmd_r2               : std_logic_vector(1 DOWNTO 0);
   SIGNAL cmd_bank_r2              : std_logic_vector(DEV_BA_WIDTH-1 DOWNTO 0);
   SIGNAL cmd_pipe_stalled         : std_logic;

   SIGNAL read_req2                : std_logic;
   SIGNAL write_req2               : std_logic;
   SIGNAL rd_grant                 : std_logic;
   SIGNAL wr_grant                 : std_logic;
   SIGNAL refr_grant               : std_logic;
   SIGNAL wr_grant_r1              : std_logic;
   SIGNAL wr_grant_r2              : std_logic;
   SIGNAL wr_grant_r3              : std_logic;
   SIGNAL prev_is_wr_grant         : std_logic;
   SIGNAL wd_fifo_rd_sreg          : std_logic_vector(9 DOWNTO 0);

   SIGNAL trc_ok_r1                : std_logic;
   SIGNAL trc_ok_r2                : std_logic;
   SIGNAL trc_ok                   : std_logic;
   SIGNAL bank_trc_ok              : std_logic_vector(7 DOWNTO 0);
   SIGNAL trc_cnt_clr              : std_logic_vector(7 DOWNTO 0);
   SIGNAL trc_cnt0                 : std_logic_vector(3 DOWNTO 0);
   SIGNAL trc_cnt1                 : std_logic_vector(3 DOWNTO 0);
   SIGNAL trc_cnt2                 : std_logic_vector(3 DOWNTO 0);
   SIGNAL trc_cnt3                 : std_logic_vector(3 DOWNTO 0);
   SIGNAL trc_cnt4                 : std_logic_vector(3 DOWNTO 0);
   SIGNAL trc_cnt5                 : std_logic_vector(3 DOWNTO 0);
   SIGNAL trc_cnt6                 : std_logic_vector(3 DOWNTO 0);
   SIGNAL trc_cnt7                 : std_logic_vector(3 DOWNTO 0);
   SIGNAL conf_trc_int             : std_logic_vector(3 DOWNTO 0);
   SIGNAL burst_cnt                : std_logic_vector(1 DOWNTO 0);
   SIGNAL burst_allow              : std_logic;

   SIGNAL ctlWdfRdEn_int           : std_logic;


BEGIN
   cs_io <= (OTHERS => 'Z');
   -- -----------------------------------------------------------------------------
   -- Output assignments
   -- -----------------------------------------------------------------------------
   ctlWdfRdEn  <= ctlWdfRdEn_int;
   cmdFifoRdEn <= load_cmd1;


   -- -----------------------------------------------------------------------------
   -- Internal reset signal, generated from reset config or Hard
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         rst <= rstConfig OR rstHard;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Signal conf_trc_int is different from confTrc because of the 
   -- pipeline used for checking the tRC violation
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
           conf_trc_int <= STD_LOGIC_VECTOR(UNSIGNED(confTrc) - 3);
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- The 1st stage of the command pipeline
   -- The values will only be updated when load_cmd1 is '1', 
   -- which is used as the command FIFO read enable signal cmdFifoRdEn
   -- -----------------------------------------------------------------------------
   cmd_cmd_r1  <= cmdFifoOut(25 DOWNTO 24);
--   cmd_addr_r1 <= cmdFifoOut(23 DOWNTO 3);
   cmd_addr_r1 <= cmdFifoOut((DEV_AD_WIDTH+2) DOWNTO 3);  -- (21:3) for x36 (DEV_AD_WIDTH=19); (22:3) for x18 (DEV_AD_WIDTH=20); (23:3) for x9  (DEV_AD_WIDTH=21)
   cmd_bank_r1 <= cmdFifoOut(2 DOWNTO 0);


   -- -----------------------------------------------------------------------------
   -- The 2nd stage of the command pipeline
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            cmd_cmd_r2  <= "00";
            cmd_addr_r2 <= (OTHERS => '0');
            cmd_bank_r2 <= (OTHERS => '0');
         ELSE
            IF ( load_cmd2 = '1' ) THEN
               cmd_cmd_r2  <= cmd_cmd_r1;
               cmd_addr_r2 <= cmd_addr_r1;
               cmd_bank_r2 <= cmd_bank_r1;
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- The command pipe will be stalled everytime trc_ok is 0
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            cmd_pipe_stalled <= '0';
         ELSE
            cmd_pipe_stalled <= NOT trc_ok;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Flags that the trc counter for a bank should be cleared
   -- -----------------------------------------------------------------------------
   trc_cnt_clr(0) <= rd_grant OR wr_grant WHEN ( cmd_bank_r2(2 DOWNTO 0) = "000" ) else '0';  -- bank 0
   trc_cnt_clr(1) <= rd_grant OR wr_grant WHEN ( cmd_bank_r2(2 DOWNTO 0) = "001" ) else '0';  -- bank 1
   trc_cnt_clr(2) <= rd_grant OR wr_grant WHEN ( cmd_bank_r2(2 DOWNTO 0) = "010" ) else '0';  -- bank 2
   trc_cnt_clr(3) <= rd_grant OR wr_grant WHEN ( cmd_bank_r2(2 DOWNTO 0) = "011" ) else '0';  -- bank 3
   trc_cnt_clr(4) <= rd_grant OR wr_grant WHEN ( cmd_bank_r2(2 DOWNTO 0) = "100" ) else '0';  -- bank 4
   trc_cnt_clr(5) <= rd_grant OR wr_grant WHEN ( cmd_bank_r2(2 DOWNTO 0) = "101" ) else '0';  -- bank 5
   trc_cnt_clr(6) <= rd_grant OR wr_grant WHEN ( cmd_bank_r2(2 DOWNTO 0) = "110" ) else '0';  -- bank 6
   trc_cnt_clr(7) <= rd_grant OR wr_grant WHEN ( cmd_bank_r2(2 DOWNTO 0) = "111" ) else '0';  -- bank 7

   -- -----------------------------------------------------------------------------
   -- Keep track the number of cycles since last access of a bank.
   -- Stop counting once the counter reaches 8 as the biggest tRC is 8.
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         -- Bank 0
         IF ( rstHard = '1' OR trc_cnt_clr(0) = '1' ) THEN
            trc_cnt0       <= "0000";
            bank_trc_ok(0) <= '0';
         ELSE
            IF ( trc_cnt0(3) = '1' ) THEN
               trc_cnt0 <= trc_cnt0;
            ELSE
                 trc_cnt0 <= STD_LOGIC_VECTOR( UNSIGNED(trc_cnt0) + 1 );   -- increment
            END IF;
            --
            IF ( trc_cnt0 >= conf_trc_int ) THEN
               bank_trc_ok(0) <= '1';
            ELSE
               bank_trc_ok(0) <= '0';
            END IF;
         END IF;

         -- Bank 1
         IF ( rstHard = '1' OR trc_cnt_clr(1) = '1' ) THEN
            trc_cnt1       <= "0000";
            bank_trc_ok(1) <= '0';
         ELSE
            IF ( trc_cnt1(3) = '1' ) THEN
               trc_cnt1 <= trc_cnt1;
            ELSE
                 trc_cnt1 <= STD_LOGIC_VECTOR( UNSIGNED(trc_cnt1) + 1 );   -- increment
            END IF;
            --
            IF ( trc_cnt1 >= conf_trc_int ) THEN
               bank_trc_ok(1) <= '1';
            ELSE
               bank_trc_ok(1) <= '0';
            END IF;
         END IF;

         -- Bank 2
         IF ( rstHard = '1' OR trc_cnt_clr(2) = '1' ) THEN
            trc_cnt2       <= "0000";
            bank_trc_ok(2) <= '0';
         ELSE
            IF ( trc_cnt2(3) = '1' ) THEN
               trc_cnt2 <= trc_cnt2;
            ELSE
               trc_cnt2 <= STD_LOGIC_VECTOR( UNSIGNED(trc_cnt2) + 1 );   -- increment
            END IF;
            --
            IF ( trc_cnt2 >= conf_trc_int ) THEN
               bank_trc_ok(2) <= '1';
            ELSE
               bank_trc_ok(2) <= '0';
            END IF;
         END IF;

         -- Bank 3
         IF ( rstHard = '1' OR trc_cnt_clr(3) = '1' ) THEN
            trc_cnt3       <= "0000";
            bank_trc_ok(3) <= '0';
         ELSE
            IF ( trc_cnt3(3) = '1' ) THEN
               trc_cnt3 <= trc_cnt3;
            ELSE
               trc_cnt3 <= STD_LOGIC_VECTOR( UNSIGNED(trc_cnt3) + 1 );   -- increment
            END IF;
            --
            IF ( trc_cnt3 >= conf_trc_int ) THEN
               bank_trc_ok(3) <= '1';
            ELSE
               bank_trc_ok(3) <= '0';
            END IF;
         END IF;

         -- Bank 4
         IF ( rstHard = '1' OR trc_cnt_clr(4) = '1' ) THEN
            trc_cnt4       <= "0000";
            bank_trc_ok(4) <= '0';
         ELSE
            IF ( trc_cnt4(3) = '1' ) THEN
               trc_cnt4 <= trc_cnt4;
            ELSE
               trc_cnt4 <= STD_LOGIC_VECTOR( UNSIGNED(trc_cnt4) + 1 );   -- increment
            END IF;
            --
            IF ( trc_cnt4 >= conf_trc_int ) THEN
               bank_trc_ok(4) <= '1';
            ELSE
               bank_trc_ok(4) <= '0';
            END IF;
         END IF;

         -- Bank 5
         IF ( rstHard = '1' OR trc_cnt_clr(5) = '1' ) THEN
            trc_cnt5       <= "0000";
            bank_trc_ok(5) <= '0';
         ELSE
            IF ( trc_cnt5(3) = '1' ) THEN
               trc_cnt5 <= trc_cnt5;
            ELSE
               trc_cnt5 <= STD_LOGIC_VECTOR( UNSIGNED(trc_cnt5) + 1 );   -- increment
            END IF;
            --
            IF ( trc_cnt5 >= conf_trc_int ) THEN
               bank_trc_ok(5) <= '1';
            ELSE
               bank_trc_ok(5) <= '0';
            END IF;
         END IF;

         -- Bank 6
         IF ( rstHard = '1' OR trc_cnt_clr(6) = '1' ) THEN
            trc_cnt6       <= "0000";
            bank_trc_ok(6) <= '0';
         ELSE
            IF ( trc_cnt6(3) = '1' ) THEN
               trc_cnt6 <= trc_cnt6;
            ELSE
               trc_cnt6 <= STD_LOGIC_VECTOR( UNSIGNED(trc_cnt6) + 1 );   -- increment
            END IF;
            --
            IF ( trc_cnt6 >= conf_trc_int ) THEN
               bank_trc_ok(6) <= '1';
            ELSE
               bank_trc_ok(6) <= '0';
            END IF;
         END IF;

         -- Bank 7
         IF ( rstHard = '1' OR trc_cnt_clr(7) = '1' ) THEN
            trc_cnt7       <= "0000";
            bank_trc_ok(7) <= '0';
         ELSE
            IF ( trc_cnt7(3) = '1' ) THEN
               trc_cnt7 <= trc_cnt7;
            ELSE
               trc_cnt7 <= STD_LOGIC_VECTOR( UNSIGNED(trc_cnt7) + 1 );   -- increment
            END IF;
            --
            IF ( trc_cnt7 >= conf_trc_int ) THEN
               bank_trc_ok(7) <= '1';
            ELSE
               bank_trc_ok(7) <= '0';
            END IF;
         END IF;

      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Signal trc_ok_r1 indicates there is no tRC violation on the 1st stage of
   -- the cmd pipeline.  It is registered, so it's timing lines up with the 
   -- 2nd stage.  It is basically a lookahead of tRC violation to achieve better
   -- timing in a continous command stream.
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rstHard = '1' ) THEN
            trc_ok_r1 <= '0';
         ELSE
            CASE cmd_bank_r1(2 DOWNTO 0) IS
               WHEN "000"  =>  trc_ok_r1 <= NOT trc_cnt_clr(0) AND bank_trc_ok(0);
               WHEN "001"  =>  trc_ok_r1 <= NOT trc_cnt_clr(1) AND bank_trc_ok(1);
               WHEN "010"  =>  trc_ok_r1 <= NOT trc_cnt_clr(2) AND bank_trc_ok(2);
               WHEN "011"  =>  trc_ok_r1 <= NOT trc_cnt_clr(3) AND bank_trc_ok(3);
               WHEN "100"  =>  trc_ok_r1 <= NOT trc_cnt_clr(4) AND bank_trc_ok(4);
               WHEN "101"  =>  trc_ok_r1 <= NOT trc_cnt_clr(5) AND bank_trc_ok(5);
               WHEN "110"  =>  trc_ok_r1 <= NOT trc_cnt_clr(6) AND bank_trc_ok(6);
               WHEN "111"  =>  trc_ok_r1 <= NOT trc_cnt_clr(7) AND bank_trc_ok(7);
               --  all cases covered, no default required
               WHEN OTHERS =>  NULL;
            END CASE;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Check tRC violation on the 2nd stage
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rstHard = '1' ) THEN
            trc_ok_r2 <= '0';
         ELSE
            CASE cmd_bank_r2(2 DOWNTO 0) IS
               WHEN "000"  =>  trc_ok_r2 <= bank_trc_ok(0);
               WHEN "001"  =>  trc_ok_r2 <= bank_trc_ok(1);
               WHEN "010"  =>  trc_ok_r2 <= bank_trc_ok(2);
               WHEN "011"  =>  trc_ok_r2 <= bank_trc_ok(3);
               WHEN "100"  =>  trc_ok_r2 <= bank_trc_ok(4);
               WHEN "101"  =>  trc_ok_r2 <= bank_trc_ok(5);
               WHEN "110"  =>  trc_ok_r2 <= bank_trc_ok(6);
               WHEN "111"  =>  trc_ok_r2 <= bank_trc_ok(7);
               --  all cases covered, no default required
               WHEN OTHERS =>  NULL;
            END CASE;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Signals trc_ok_r1 and trc_ok_r2 need to be used together to indicate
   -- there is no tRC violation
   -- -----------------------------------------------------------------------------
   trc_ok <= ( cmd_pipe_stalled AND trc_ok_r2 ) OR ( NOT cmd_pipe_stalled AND trc_ok_r1 );


   -- -----------------------------------------------------------------------------
   -- - Burst counter, burst_cnt
   -- - Burst_allow signal is used to control how often rd_grant or
   --   wr_grant can be issued:
   --      for BL2, burst_allow is set all the time
   --      for BL4, burst_allow is set every 2 cycles
   --      for BL8, burst allow is set every 4 cycles
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            burst_cnt   <= "00";
            burst_allow <= '0';
         ELSE
            burst_cnt <= STD_LOGIC_VECTOR( UNSIGNED(burst_cnt) + 1 );   -- increment

            IF ( confBL = "0010" ) THEN               -- for BL2
               burst_allow <= '1';
            ELSE
               IF ( confBL = "0100" ) THEN            -- for BL4
                  burst_allow <= burst_cnt(0);
               ELSE                                   -- for BL8
                  burst_allow <= AND_REDUCE( burst_cnt );
               END IF;
            END IF;
         END IF;
      END IF;
   END PROCESS;

   read_req2        <= NOT cmd_cmd_r2(1) AND     cmd_cmd_r2(0);  -- set when cmd_cmd_r2 = "01"
   write_req2       <= NOT cmd_cmd_r2(1) AND NOT cmd_cmd_r2(0);  -- set when cmd_cmd_r2 = "00"
   prev_is_wr_grant <= wd_fifo_rd_sreg(0);


   -- -----------------------------------------------------------------------------
   -- Controller state machine (FSM)
   -- It only processes the commands from the 2nd stage of the command pipeline.
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rstHard = '1' ) THEN
            ctl_state <= CTL_IDLE;
         ELSE
            CASE ctl_state IS
               -- ----------------------------------------------------------------
               WHEN CTL_IDLE =>
               -- ----------------------------------------------------------------
                  -- AREF request is only checked in CTL_IDLE to simplify logic.
                  -- The AREF request will be responded in maximum 4 cycles 
                  -- depending the current state.
                  IF ( refr_req = '1' ) THEN
                     ctl_state <= CTL_PROC_REFR;
                  ELSE
                     IF ( NOT cmdFifoEmpty = '1' ) THEN
                        ctl_state <= CTL_LOAD_CMD1;
                     ELSE
                        ctl_state <= CTL_IDLE;
                     END IF;
                  END IF;

               -- ----------------------------------------------------------------
               WHEN CTL_LOAD_CMD1 =>
               -- ----------------------------------------------------------------
                  ctl_state <= CTL_LOAD_CMD2;

               -- ----------------------------------------------------------------
               WHEN CTL_LOAD_CMD2 =>
               -- ----------------------------------------------------------------
                  -- If the cmd FIFO is empty, there is no more command to load,
                  -- so go to CTL_PROC_LAST_CMD and process the only command
                  -- in the pipeline.
                  IF ( NOT cmdFifoEmpty = '1' ) THEN
                     ctl_state <= CTL_PROC_CMD;
                  ELSE
                     ctl_state <= CTL_PROC_LAST_CMD;
                  END IF;

               -- ----------------------------------------------------------------
               WHEN CTL_PROC_CMD =>
               -- ----------------------------------------------------------------
                  -- When the cmd FIFO is empty or AREF request is high, finish 
                  -- processing the current cmd, then go to CTL_PROC_LAST_CMD 
                  -- to process the last cmd in the pipeline.
                  -- Otherwise stay in this state.
                  IF ( ( cmdFifoEmpty = '1' OR refr_req = '1' ) AND
                       ( write_req2 = '1' OR ( read_req2 = '1' AND NOT prev_is_wr_grant = '1') ) AND trc_ok = '1' AND burst_allow = '1' ) THEN
                     ctl_state <= CTL_PROC_LAST_CMD;
                  ELSE
                     ctl_state <= CTL_PROC_CMD;
                  END IF;

               -- ----------------------------------------------------------------
               WHEN CTL_PROC_LAST_CMD =>
               -- ----------------------------------------------------------------
                  -- Processing the last command.   The command pipeline will be
                  -- empty after this state.  Go back to CTL_IDLE state after
                  -- rd_grant or wr_grant is issued
                  IF ( ( write_req2 = '1' OR ( read_req2 = '1' AND NOT prev_is_wr_grant = '1') ) AND trc_ok = '1' AND burst_allow = '1' ) THEN
                     ctl_state <= CTL_IDLE;
                  ELSE
                     ctl_state <= CTL_PROC_LAST_CMD;
                  END IF;

               -- ----------------------------------------------------------------
               WHEN CTL_PROC_REFR =>
               -- ----------------------------------------------------------------
                  IF ( refr_done = '1' ) THEN
                     ctl_state <= CTL_IDLE;
                  ELSE
                     ctl_state <= CTL_PROC_REFR;
                  END IF;

               -- ----------------------------------------------------------------
               WHEN OTHERS  =>
               -- ----------------------------------------------------------------
                  ctl_state <= CTL_IDLE;
            END CASE;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Combinatorial logic for the "ctl_state" Controller FSM
   -- -----------------------------------------------------------------------------
   PROCESS (ctl_state, cmdFifoEmpty, read_req2, prev_is_wr_grant, trc_ok, burst_allow, write_req2, refr_req, refr_done, refr_trc_ok)
   BEGIN
      -- Default assignments.
      -- Only the outputs that are different from the default values will be updated.
      -- This saves typing and makes the code more readable.
      load_cmd1    <= '0';
      load_cmd2    <= '0';
      rd_grant     <= '0';
      wr_grant     <= '0';
      refr_grant   <= '0';
      refr_req_clr <= '0';

      CASE ctl_state IS
         -----------------------------------------------------------------------
         WHEN CTL_IDLE =>
         -----------------------------------------------------------------------
            -- Use default assignments

         -----------------------------------------------------------------------
         WHEN CTL_LOAD_CMD1 =>
         -----------------------------------------------------------------------
            -- Load the 1st stage
            load_cmd1 <= '1';

         -----------------------------------------------------------------------
         WHEN CTL_LOAD_CMD2 =>
         -----------------------------------------------------------------------
            IF ( NOT cmdFifoEmpty = '1' ) THEN
               -- If cmd FIFO is not empty, 
               -- load both the 1st stage and 2nd stage
               load_cmd1 <= '1';
               load_cmd2 <= '1';
            ELSE
               -- Otherwise, don't load the 1st stage and 
               -- push the 1st stage to 2nd for processing
               load_cmd1 <= '0';
               load_cmd2 <= '1';
            END IF;

         -----------------------------------------------------------------------
         WHEN CTL_PROC_CMD =>
         -----------------------------------------------------------------------
            -- When in CTL_PROC_CMD state, it will always stay in this state
            -- unless a rd_grant or wr_grant is issued.

            -- One NOP needs to be inserted when transitioning from write to read.           
            IF ( read_req2 AND NOT prev_is_wr_grant AND trc_ok AND burst_allow ) = '1' THEN
               rd_grant <= '1';
            ELSE
               IF ( write_req2 = '1' AND trc_ok = '1' AND burst_allow = '1' ) THEN
                  wr_grant <= '1';
               END IF;
            END IF;

            IF ( ( cmdFifoEmpty = '1' OR refr_req = '1') AND
                 ( write_req2 = '1' OR ( read_req2 = '1' AND NOT prev_is_wr_grant = '1' ) ) AND trc_ok = '1' AND burst_allow = '1' ) THEN
               -- Going to CTL_PROC_LAST_CMD state;only load the 2nd stage
               load_cmd1 <= '0';
               load_cmd2 <= '1';
            ELSE
               IF ( ( write_req2 = '1' OR ( read_req2 = '1' AND NOT prev_is_wr_grant = '1' ) ) AND trc_ok = '1' AND burst_allow = '1' ) THEN
                  -- Staying in CTL_PROC_CMD state; load both 1st and 2nd stage
                  load_cmd1 <= '1';
                  load_cmd2 <= '1';
               END IF;
            END IF;

         -----------------------------------------------------------------------
         WHEN CTL_PROC_LAST_CMD =>
         -----------------------------------------------------------------------
            -- Finish processing the last command in the pipeline
            IF ( read_req2 = '1' AND NOT prev_is_wr_grant = '1' AND trc_ok = '1' AND burst_allow = '1' ) THEN
               rd_grant <= '1';
            ELSE
               IF ( write_req2 = '1' AND trc_ok = '1' AND burst_allow = '1' ) THEN
                  wr_grant <= '1';
               END IF;
            END IF;

         -----------------------------------------------------------------------
         WHEN CTL_PROC_REFR =>
         -----------------------------------------------------------------------
            -- Processing AREF request
            refr_req_clr <= '1';

            IF ( refr_done = '1' ) THEN
               refr_grant <= '0';
            ELSE
               IF ( refr_trc_ok = '1' ) THEN
                  refr_grant <= '1';
               END IF;
            END IF;

         -----------------------------------------------------------------------
         WHEN OTHERS  =>
         -----------------------------------------------------------------------
            -- Use default assignments

      END CASE;
   END PROCESS;

   -- -----------------------------------------------------------------------------
   -- -----------------------------------------------------------------------------
   ctlIoWdEn    <= ctlWdfRdEn_int;
   rldWriteDone <= ctlWdfRdEn_int;


   -- -----------------------------------------------------------------------------
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         rd_grant_sreg <= rd_grant_sreg(13 DOWNTO 0) & rd_grant;  -- shift in "rd_grant"
         -- This signal controls when the deskew tap adjustment can be done.
         -- Avoid adjusting when data could be coming in - both earliest data 
         -- arrival and latest data arrival is taken care of.
         ctlOkEdge <= NOT OR_REDUCE( rd_grant_sreg );
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Generate the command for the RLDRAM
   -- ** Note :
   --    refUser command is not supported at this time (as of 09/07/2005)
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rstHard = '1' ) THEN
            ctlCmdOut  <= CMD_DESEL;
            ctlAddrOut <= (OTHERS => '0');
            ctlBankOut <= (OTHERS => '0');
         ELSE
            IF ( rstConfig = '1' ) THEN
               ctlCmdOut  <= rstCmd;
               ctlAddrOut <= rstAd;
               ctlBankOut <= rstBa;
            ELSE
               IF ( rd_grant = '1' ) THEN
                  ctlCmdOut  <= CMD_RD;
                  ctlAddrOut <= cmd_addr_r2;
                  ctlBankOut <= cmd_bank_r2;
               ELSE
                  IF ( wr_grant = '1' ) THEN
                     ctlCmdOut  <= CMD_WR;
                     ctlAddrOut <= cmd_addr_r2;
                     ctlBankOut <= cmd_bank_r2;
                  ELSE
                     IF ( refr_grant = '1' ) THEN
                        ctlCmdOut  <= CMD_REF;
                        ctlBankOut <= refr_bank;
--                     ELSE
--                        IF ( refUser = '1' ) THEN
--                           ctlCmdOut  <= CMD_REF;
--                           ctlBankOut <= cmd_bank_r2;
                     ELSE
                        ctlCmdOut <= CMD_NOP;
                     END IF;
                  END IF;
               END IF;
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Refresh counter
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( refr_done = '1' OR rst = '1' ) THEN
            refr_cnt <= confRcCnt1(3 DOWNTO 0) & confRcCnt0;  -- 12-bit value
         ELSE
            refr_cnt <= STD_LOGIC_VECTOR( UNSIGNED(refr_cnt) - 1 );   -- decrement
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Refresh logic
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            refr_req           <= '0';
            refr_done          <= '0';
            refr_req_clr_sreg  <= (others => '0');
            refr_trc_ok        <= '0';
            refr_bank          <= "000";
         ELSE
            IF ( refr_req_clr = '1' ) THEN
               refr_req <= '0';
            ELSE
               IF ( ( refr_cnt = "0000" & confRcMargin ) AND confAutoRefrEn = '1' ) THEN
                  refr_req <= '1';
               END IF;
            END IF;

            refr_req_clr_sreg <= refr_req_clr_sreg(6 DOWNTO 0) & refr_req_clr;  -- shift in "refr_req_clr"

            -- Signal refr_trc_ok indicates tRC is satisfied in the 
            -- worst case scenario and refr_grant can be issued
            IF ( confTrc = "0100" ) THEN                        -- when confTrc = 4
               refr_trc_ok <= refr_req_clr_sreg(1);
            ELSE
               IF ( confTrc = "0110" ) THEN                     -- when confTrc = 6
                  refr_trc_ok <= refr_req_clr_sreg(3);
               ELSE
                  refr_trc_ok <= refr_req_clr_sreg(5);
               END IF;
            END IF;

            -- Finish refreshing all 8 banks
            refr_done <= refr_grant AND (refr_bank(2) AND refr_bank(1) AND refr_bank(0));  -- when refr_bank="111"

            IF ( NOT refr_grant = '1' ) THEN
               refr_bank <= "000";
            ELSE
               CASE refr_bank IS
                  WHEN "000"  =>  refr_bank <= "001";
                  WHEN "001"  =>  refr_bank <= "010";
                  WHEN "010"  =>  refr_bank <= "011";
                  WHEN "011"  =>  refr_bank <= "100";
                  WHEN "100"  =>  refr_bank <= "101";
                  WHEN "101"  =>  refr_bank <= "110";
                  WHEN "110"  =>  refr_bank <= "111";
                  WHEN "111"  =>  refr_bank <= "000";
                  --  all cases covered, no default required
                  WHEN OTHERS =>  NULL;
               END CASE;
            END IF;

         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Enable driving write data to the pads at the appropriate time.
   -- Insert wr_grant into the right location based on burst length and write latency.
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            wr_grant_r1     <= '0';
            wr_grant_r2     <= '0';
            wr_grant_r3     <= '0';
            wd_fifo_rd_sreg <= (others => '0');
            ctlWdfRdEn_int  <= '0';
         ELSE
            wr_grant_r1 <= wr_grant;
            wr_grant_r2 <= wr_grant_r1;
            wr_grant_r3 <= wr_grant_r2;

            IF ( confBL = "1000" ) THEN               -- for BL8
               -- need to extend the pulse width to 4
               wd_fifo_rd_sreg(0) <= wr_grant OR wr_grant_r1 OR wr_grant_r2 OR wr_grant_r3;
            ELSE
               IF ( confBL = "0100" ) THEN              -- for BL4
                  -- need to extend the pulse width to 2
                  wd_fifo_rd_sreg(0) <= wr_grant OR wr_grant_r1;
               ELSE
                  wd_fifo_rd_sreg(0) <= wr_grant;     -- for BL2, one pulse
               END IF;
            END IF;

            wd_fifo_rd_sreg(9 DOWNTO 1) <= wd_fifo_rd_sreg(8 DOWNTO 0);

            IF ( confWL = "0101" ) THEN           -- confWL = 5
               ctlWdfRdEn_int <= wd_fifo_rd_sreg(4);
            ELSE
               IF ( confWL = "0111" ) THEN           -- confWL = 7
                  ctlWdfRdEn_int <= wd_fifo_rd_sreg(6);
               ELSE
                  ctlWdfRdEn_int <= wd_fifo_rd_sreg(8);
               END IF;
            END IF;
         END IF;
      END IF;
   END PROCESS;


END ARCHITECTURE rtl;
