--*****************************************************************************
-- DISCLAIMER OF LIABILITY
-- 
-- This text/file contains proprietary, confidential
-- information of Xilinx, Inc., is distributed under license
-- from Xilinx, Inc., and may be used, copied and/or
-- disclosed only pursuant to the terms of a valid license
-- agreement with Xilinx, Inc. Xilinx hereby grants you a 
-- license to use this text/file solely for design, simulation, 
-- implementation and creation of design files limited 
-- to Xilinx devices or technologies. Use with non-Xilinx 
-- devices or technologies is expressly prohibited and 
-- immediately terminates your license unless covered by
-- a separate agreement.
--
-- Xilinx is providing this design, code, or information 
-- "as-is" solely for use in developing programs and 
-- solutions for Xilinx devices, with no obligation on the 
-- part of Xilinx to provide support. By providing this design, 
-- code, or information as one possible implementation of 
-- this feature, application or standard, Xilinx is making no 
-- representation that this implementation is free from any 
-- claims of infringement. You are responsible for 
-- obtaining any rights you may require for your implementation. 
-- Xilinx expressly disclaims any warranty whatsoever with 
-- respect to the adequacy of the implementation, including 
-- but not limited to any warranties or representations that this
-- implementation is free from claims of infringement, implied 
-- warranties of merchantability or fitness for a particular 
-- purpose.
--
-- Xilinx products are not intended for use in life support
-- appliances, devices, or systems. Use in such applications is
-- expressly prohibited.
--
-- Any modifications that are made to the Source Code are 
-- done at the user's sole risk and will be unsupported.
--
-- Copyright (c) 2007 Xilinx, Inc. All rights reserved.
--
-- This copyright and support notice must be retained as part 
-- of this text at all times. 
--*****************************************************************************
--   ____  ____
--  /   /\/   /
-- /___/  \  /    Vendor: Xilinx
-- \   \   \/     Version: 1.1
--  \   \         Filename: rld_dly_cal_sm.vhd
--  /   /         Timestamp: 24 August 2007
-- /___/   /\     
-- \   \  /  \
--  \___\/\___\
--
--
--	Device:  Virtex-5
--	Purpose: Calibrates the IDELAY tap values for the DQ and QK inputs to allow 
--		capture of the read data into the FPGA clock domain and determines
--		when data is valid using the read enable provided by the memory.
-- Revision History:
--   Rev 1.0 - initial release
--   Rev 1.1 - added RL_CK_PERIOD parameter, change to allow more taps for 
--             2nd stage calibration. Paramaterized read data signals to allow
--             clock to data ratios of 1:9 or 1:18, added the X_CORE synthesis 
--             attribute
--*****************************************************************************

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;


ENTITY rld_dly_cal_sm IS
   GENERIC (
      DEV_DQ_WIDTH            : integer := 18;     -- data width of the memory device
      RL_CK_PERIOD            : integer := 3003;   -- CK clock period of the RLDRAM in ps      
      QK_DATA_WIDTH           : integer := 9
   );
   PORT (
      clk                : IN  std_logic;   -- clock270
      rst                : IN  std_logic;   -- rst or Issue calibration signal
      start_cal          : IN  std_logic;   -- flag to indicate I can start calibration
      second_stage       : IN  std_logic;   -- look for 2nd stage data calibration
      qvld_cal           : IN  std_logic;   -- align the QVLD when active high
      wait_for_refresh   : IN  std_logic;   -- wait while refresh occurs
      rd_data_rise       : IN  std_logic_vector(QK_DATA_WIDTH-1 DOWNTO 0);   -- read data rise from ISERDES
      rd_data_fall       : IN  std_logic_vector(QK_DATA_WIDTH-1 DOWNTO 0);   -- read data fall from ISERDES
      rd_data_valid      : IN  std_logic;   -- QVLD from ISERDES
      rd_data_valid_r    : IN  std_logic;   -- QVLD registered 1 clock cycle

      dq_dly_rst         : OUT std_logic;   -- Data (DQ) IDELAY controls
      dq_dly_ce          : OUT std_logic;  
      dq_dly_inc         : OUT std_logic;  

      qk_dly_rst         : OUT std_logic;   -- Strobe (QK) IDELAY controls
      qk_dly_ce          : OUT std_logic;  
      qk_dly_inc         : OUT std_logic;  

      first_cal_done     : OUT std_logic;   -- first stage of calibration is done
      dly_cal_done       : OUT std_logic;   -- calibration is done
      delay_select       : OUT std_logic_vector(1 DOWNTO 0);   -- this indicates data early, aligned, or late with respect to QVLD

      QK_tap_count       : OUT std_logic_vector(5 DOWNTO 0);   -- tap count output of Strobe
      DQ_tap_count       : OUT std_logic_vector(5 DOWNTO 0);   -- tap count output of Data
      test_cal           : OUT std_logic_vector(127 DOWNTO 0)  --  Debug port
   );
   ATTRIBUTE X_CORE_INFO : STRING;
   ATTRIBUTE X_CORE_INFO of rld_dly_cal_sm : ENTITY IS "rld_dly_cal_sm, V5_rldram2_phy_vhdl"; 
END ENTITY rld_dly_cal_sm;


ARCHITECTURE rtl OF rld_dly_cal_sm IS

   function CC (COND_EXPR:INTEGER; COND_VAL:INTEGER; IF_VAL:std_logic_vector; THEN_VAL:std_logic_vector) return std_logic_vector is 
   begin 
     if COND_EXPR = COND_VAL then
       return IF_VAL; 
     else
       return THEN_VAL; 
     end if; 
   end;

   -- defines counts and window size for calibration
   CONSTANT DQ_TAP_END   : std_logic_vector(5 DOWNTO 0) := "111100";  -- (60 decimal)
   CONSTANT QK_TAP_END   : std_logic_vector(5 DOWNTO 0) := "111111";  -- (63 decimal)
   CONSTANT WINDOW_SIZE  : std_logic_vector(5 DOWNTO 0) := "001100";  -- (12 decimal)
   CONSTANT WINDOW_SIZE2 : std_logic_vector(5 DOWNTO 0) := CC(DEV_DQ_WIDTH, 9, "001111", "010100"); --15 decimal or 20 decimal
   CONSTANT WINDOW_SIZE3 : std_logic_vector(5 DOWNTO 0) := "010100";  -- (20 decimal)

   -- defines flag values for "delay select"
   -- data early, aligned, or late with respect to QVLD
   CONSTANT ALIGNED  : std_logic_vector(1 DOWNTO 0) := "00";
   CONSTANT EARLY    : std_logic_vector(1 DOWNTO 0) := "01";
   CONSTANT LATE     : std_logic_vector(1 DOWNTO 0) := "10";

   -- patterns used for write data
   CONSTANT data1  : std_logic_vector(QK_DATA_WIDTH-1 DOWNTO 0) := (OTHERS => '1');  -- is 0x1FF, 1_1111_1111  (rising edge data)
   CONSTANT data2  : std_logic_vector(QK_DATA_WIDTH-1 DOWNTO 0) := (OTHERS => '0');  -- is 0x000, 0_0000_0000  (falling edge data)
   CONSTANT data3  : std_logic_vector(QK_DATA_WIDTH-1 DOWNTO 0) := CC(QK_DATA_WIDTH, 9, "010101010", "010101010010101010");  -- is 0x0AA or 0x154AA (rising edge data)
   CONSTANT data4  : std_logic_vector(QK_DATA_WIDTH-1 DOWNTO 0) := CC(QK_DATA_WIDTH, 9, "101010101", "101010101101010101");  -- is 0x155 or 0x2ac55 (falling edge data)

   -- First write data  : { data1, data2 } = { 1_1111_1111, 0_0000_0000 } = "3FE00" 
   -- Second write data : { data3, data4 } = { 0_1010_1010, 1_0101_0101 } = "15555"  


   SIGNAL dly_rst                  : std_logic;
   SIGNAL dly_inc                  : std_logic;
   SIGNAL dly_ce                   : std_logic;

   SIGNAL cal_begin                : std_logic;
   SIGNAL start_cal_r              : std_logic_vector(5 DOWNTO 0);

   SIGNAL qk_flag                  : std_logic;
   SIGNAL dq_flag                  : std_logic;

   SIGNAL tap_cmd                  : std_logic_vector(1 DOWNTO 0);

   SIGNAL tap_wait_en              : std_logic;
   SIGNAL tap_wait_cnt             : std_logic_vector(2 DOWNTO 0);

   SIGNAL dq_error                 : std_logic;
   SIGNAL dq_error2                : std_logic;

   SIGNAL first_edge_found         : std_logic;
   SIGNAL second_edge_found        : std_logic;
   SIGNAL rst_flags                : std_logic;
   SIGNAL rst_flags_r              : std_logic;
   SIGNAL set_center               : std_logic_vector(2 DOWNTO 0);

   SIGNAL dly_rst_r                : std_logic;
   SIGNAL end_of_taps              : std_logic;
   SIGNAL DQ_bigger                : std_logic;
   SIGNAL DQ_bigger_r              : std_logic;
   SIGNAL QK_bigger                : std_logic;
   SIGNAL QK_bigger_r              : std_logic;
   SIGNAL valid_found              : std_logic;

   SIGNAL DQ_centered              : std_logic;
   SIGNAL QK_centered              : std_logic;
   SIGNAL DQ_centered_r            : std_logic;
   SIGNAL QK_centered_r            : std_logic;

   SIGNAL first_edge_detect        : std_logic;
   SIGNAL first_edge_detect_r      : std_logic;
   SIGNAL second_edge_detect       : std_logic;
   SIGNAL second_edge_detect_r     : std_logic;

   SIGNAL QK_tap_left              : std_logic_vector(5 DOWNTO 0);
   SIGNAL QK_tap_window            : std_logic_vector(5 DOWNTO 0);
   SIGNAL QK_tap_center            : std_logic_vector(5 DOWNTO 0);

   SIGNAL DQ_tap_left              : std_logic_vector(5 DOWNTO 0);
   SIGNAL DQ_tap_window            : std_logic_vector(5 DOWNTO 0);
   SIGNAL DQ_tap_center            : std_logic_vector(5 DOWNTO 0);

   SIGNAL tap_center               : std_logic_vector(5 DOWNTO 0);
   SIGNAL tap_range                : std_logic_vector(5 DOWNTO 0);
   
   SIGNAL inc_up                   : std_logic;
   SIGNAL inc_up2                  : std_logic;

   --Working Registers
   SIGNAL QK_tap_center_WORK       : std_logic_vector(5 DOWNTO 0);
   SIGNAL DQ_tap_center_WORK       : std_logic_vector(5 DOWNTO 0);
   -- added 8/22 as a test
   SIGNAL QK_tap_center_WORK_r       : std_logic_vector(5 DOWNTO 0);
   SIGNAL DQ_tap_center_WORK_r       : std_logic_vector(5 DOWNTO 0);
   -- end of what added
   SIGNAL QK_tap_count_r           : std_logic_vector(5 DOWNTO 0);
   SIGNAL DQ_tap_count_r           : std_logic_vector(5 DOWNTO 0);

   SIGNAL tap_range_center         : std_logic_vector(5 DOWNTO 0);   -- not registered, just shifted
   SIGNAL dly_rst_w                : std_logic;
   SIGNAL rd_data                  : std_logic_vector(2*QK_DATA_WIDTH-1 DOWNTO 0);
   SIGNAL first_check_data         : std_logic_vector(2*QK_DATA_WIDTH-1 DOWNTO 0);
   SIGNAL second_check_data        : std_logic_vector(2*QK_DATA_WIDTH-1 DOWNTO 0);
   SIGNAL dq_error3                : std_logic;

   -- States for "id_ns/id_cs" state machine 
--   CONSTANT IDLE      : std_logic_vector(3 DOWNTO 0) := "0000";  -- 000
--   CONSTANT TAP_INC   : std_logic_vector(3 DOWNTO 0) := "0001";  -- 001
--   CONSTANT TAP_DEC   : std_logic_vector(3 DOWNTO 0) := "0010";  -- 002
--   CONSTANT TAP_RST   : std_logic_vector(3 DOWNTO 0) := "0100";  -- 004
--   CONSTANT TAP_WAIT  : std_logic_vector(3 DOWNTO 0) := "1000";  -- 008
   -- IDELAY state variables: states for "id_ns/id_cs" state machine 
   type idStates is ( IDLE, TAP_INC, TAP_DEC, TAP_RST, TAP_WAIT );
   SIGNAL id_ns, id_cs  : idStates;

   -- States for "f_ns/f_cs" state machine 
--   CONSTANT Reset     : std_logic_vector(11 DOWNTO 0) := "000000000000";  -- 000
--   CONSTANT QK_INC    : std_logic_vector(11 DOWNTO 0) := "000000000001";  -- 001
--   CONSTANT QK_RST    : std_logic_vector(11 DOWNTO 0) := "000000000010";  -- 002
--   CONSTANT DQ_INC    : std_logic_vector(11 DOWNTO 0) := "000000000100";  -- 004
--   CONSTANT DQ_RST    : std_logic_vector(11 DOWNTO 0) := "000000001000";  -- 008
--   CONSTANT Center    : std_logic_vector(11 DOWNTO 0) := "000000010000";  -- 010
--   CONSTANT F_Done    : std_logic_vector(11 DOWNTO 0) := "000000100000";  -- 020
--   CONSTANT Both_INC  : std_logic_vector(11 DOWNTO 0) := "000001000000";  -- 040
--   CONSTANT Both_DEC  : std_logic_vector(11 DOWNTO 0) := "000010000000";  -- 080
--   CONSTANT DQ_DEC    : std_logic_vector(11 DOWNTO 0) := "000100000000";  -- 100
--   CONSTANT DQ_INC2   : std_logic_vector(11 DOWNTO 0) := "001000000000";  -- 200
--   CONSTANT DQ_CENT   : std_logic_vector(11 DOWNTO 0) := "010000000000";  -- 400
--   CONSTANT S_Done    : std_logic_vector(11 DOWNTO 0) := "100000000000";  -- 800
   -- First SM state variables: states for "f_ns/f_cs" state machine 
   type fStates is ( Reset, QK_INC, QK_RST, DQ_INC, DQ_RST, Center, F_Done, Both_INC, Both_DEC, DQ_DEC, DQ_INC2, DQ_CENT, S_Done );
   SIGNAL f_ns, f_cs  : fStates;

   -- Command parameters
   CONSTANT TAP_CMD_NOP  : std_logic_vector(1 DOWNTO 0) := "00";
   CONSTANT TAP_CMD_RST  : std_logic_vector(1 DOWNTO 0) := "01";
   CONSTANT TAP_CMD_DEC  : std_logic_vector(1 DOWNTO 0) := "10";
   CONSTANT TAP_CMD_INC  : std_logic_vector(1 DOWNTO 0) := "11";

   -- IDELAY state variables
--   SIGNAL id_cs                    : std_logic_vector(3 DOWNTO 0);
--   SIGNAL id_ns                    : std_logic_vector(3 DOWNTO 0);

   -- First SM state variables
--   SIGNAL f_cs                     : std_logic_vector(11 DOWNTO 0);
--   SIGNAL f_ns                     : std_logic_vector(11 DOWNTO 0);

   SIGNAL dq_dly_rst_int           : std_logic;
   SIGNAL dq_dly_ce_int            : std_logic;
   SIGNAL dq_dly_inc_int           : std_logic;

   SIGNAL qk_dly_rst_int           : std_logic;
   SIGNAL qk_dly_ce_int            : std_logic;
   SIGNAL qk_dly_inc_int           : std_logic;

   -- internal counter signals, a version goes out...
   SIGNAL first_cal_done_int       : std_logic;
   SIGNAL dly_cal_done_int         : std_logic;
   SIGNAL delay_select_int         : std_logic_vector(1 DOWNTO 0);

   SIGNAL QK_tap_count_int         : std_logic_vector(5 DOWNTO 0);
   SIGNAL DQ_tap_count_int         : std_logic_vector(5 DOWNTO 0);


BEGIN
   dq_dly_rst     <= dq_dly_rst_int;
   dq_dly_ce      <= dq_dly_ce_int;
   dq_dly_inc     <= dq_dly_inc_int;

   qk_dly_rst     <= qk_dly_rst_int;
   qk_dly_ce      <= qk_dly_ce_int;
   qk_dly_inc     <= qk_dly_inc_int;

   first_cal_done <= first_cal_done_int;
   dly_cal_done   <= dly_cal_done_int;
   delay_select   <= delay_select_int;
   QK_tap_count   <= QK_tap_count_int;
   DQ_tap_count   <= DQ_tap_count_int;

   -- assign outputs for IDELAY based on flags
   qk_dly_rst_int  <= qk_flag AND dly_rst_w;  -- if qk_flag=1, then qk_dly_rst=dly_rst_w, else qk_dly_rst=0
   qk_dly_ce_int   <= qk_flag AND dly_ce;
   qk_dly_inc_int  <= qk_flag AND dly_inc;
   --
   dq_dly_rst_int  <= dq_flag AND dly_rst_w;  -- if dq_flag=1, then dq_dly_rst=dly_rst_w, else dq_dly_rst=0
   dq_dly_ce_int   <= dq_flag AND dly_ce;
   dq_dly_inc_int  <= dq_flag AND dly_inc;


   -- -----------------------------------------------------------------------------
   -- Concatenate read data { 9-bit RISE : 9-bit FALL } before comparing with patterns
   -- -----------------------------------------------------------------------------
   rd_data <= rd_data_rise & rd_data_fall;  -- 18-bit read data

   -- Concatonate the Data for checking
   first_check_data  <= data1 & data2;
   second_check_data <= data3 & data4;


   -- -----------------------------------------------------------------------------
   -- Test Signals used for debugging
   -- -----------------------------------------------------------------------------
--   test_cal(127 DOWNTO 0) <= (others=>'0');
   test_cal(127 DOWNTO 73) <= (others=>'0');
   test_cal(72 DOWNTO 0) <= 
--         valid_found,                                                        -- bit (73)
--         QK_tap_center & DQ_tap_center &                                     -- bit (72:67), bit (66:61)
         "000000000000" &                                                    -- bit (72:61)
--         QK_tap_left & DQ_tap_left &                                         -- bit (60:55), bit (54:49)
         "000000000000" &                                                    -- bit (60:49)
         QK_tap_count_int & DQ_tap_count_int &                               -- bit (48:43), bit (42:37)
         first_cal_done_int & second_stage & dly_cal_done_int &              -- bit (36,35,34)
         rst_flags & first_edge_detect & second_edge_detect & end_of_taps &  -- bit (33,32,31,30)
         wait_for_refresh & dq_error3 & dq_error2 &  dq_error &              -- bit (29,28,27,26)
--         delay_select & dq_error3 & dq_error2 & dq_error &                   -- bit (29,28,27,26)
--         tap_cmd & f_ns & f_cs &                                             -- bit (25:24), (23:12), (11:0)
--         "00" & "000000000000" & f_cs;                                       -- bit (25:24), (23:12), (11:0)
         "00" & "000000000000" & "000000000000";                             -- bit (25:24), (23:12), (11:0)


   -- -----------------------------------------------------------------------------
   -- Hold the IDELAY reset signal one extra clock cycle
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (rst = '1') THEN
            dly_rst_r <= '0';
         ELSE
            dly_rst_r <= dly_rst;
         END IF;
      END IF;
   END PROCESS;

   dly_rst_w <= dly_rst OR dly_rst_r ;


   -- -----------------------------------------------------------------------------
   -- Delay the start signal
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (rst = '1') THEN
            start_cal_r(0) <= '0';
            start_cal_r(1) <= '0';
            start_cal_r(2) <= '0';
            start_cal_r(3) <= '0';
            start_cal_r(4) <= '0';
            start_cal_r(5) <= '0';
         ELSE
            start_cal_r(0) <= start_cal;
            start_cal_r(1) <= start_cal_r(0);
            start_cal_r(2) <= start_cal_r(1);
            start_cal_r(3) <= start_cal_r(2);
            start_cal_r(4) <= start_cal_r(3);
            start_cal_r(5) <= start_cal_r(4);
         END IF;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            cal_begin <= '0';
         ELSE
            IF ( start_cal_r(4) = '1' AND NOT start_cal_r(5) = '1' ) THEN
               cal_begin <= '1';
            ELSE
               IF ( dly_inc = '1' ) THEN
                  cal_begin <= '0';
               ELSE
                  cal_begin <= cal_begin;
               END IF;
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Determine how the data lines up with the QVLD
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (rst = '1') THEN
            delay_select_int <= ALIGNED;
            valid_found      <= '0';
         ELSE
            IF ( NOT valid_found = '1' AND qvld_cal = '1' AND rd_data_valid = '1' AND NOT rd_data_valid_r = '1' ) THEN
               valid_found <= '1';
               IF ( rd_data = first_check_data ) THEN
                  delay_select_int <= ALIGNED;
               ELSE
                  IF (rd_data = second_check_data) THEN
                     delay_select_int <= EARLY;
                  ELSE
                     delay_select_int <= LATE;
                  END IF;
               END IF;
            ELSE
               delay_select_int <= delay_select_int;
               valid_found      <= valid_found;
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Update the Error signal when data is valid
   -- Check the 1st SDR data for error
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            dq_error <= '0';
         ELSE
            IF ( (tap_wait_cnt >= "110") AND NOT wait_for_refresh = '1' ) THEN
               -- only check data when IDLE to ensure settled
               IF ( rd_data = first_check_data ) THEN
                  dq_error <= '0';
               ELSE
                  dq_error <= '1';
               END IF;
            ELSE
               dq_error <= dq_error;
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Check the 2nd SDR data for error
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            dq_error2 <= '0';
         ELSE
            IF ( (tap_wait_cnt >= "110") AND NOT wait_for_refresh = '1' ) THEN
               IF (rd_data = second_check_data) THEN
                  dq_error2 <= '0';
               ELSE
                  dq_error2 <= '1';
               END IF;
            ELSE
               dq_error2 <= dq_error2;   -- hold the previous value
            END IF;
         END IF;
      END IF;
   END PROCESS;

   -- if second_stage is '1', then dq_error3=(dq_error AND dq_error2), else dq_error3=dq_error
   dq_error3 <= ((dq_error AND dq_error2) AND second_stage) OR (dq_error AND NOT second_stage);


   -- -----------------------------------------------------------------------------
   -- State machine "id_cs"
   --    Controlling the QK and DQ IDELAYs
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            id_cs <= TAP_RST;   -- reset tap count, works for system reset and re-calibration
         ELSE
            id_cs <= id_ns;
         END IF;
      END IF;
   END PROCESS;

   -- combinatorial logic for "id_ns" state machine
   PROCESS (id_cs, tap_cmd, tap_wait_cnt)
   BEGIN
      dly_rst     <= '0';
      dly_ce      <= '0';
      dly_inc     <= '0';
      tap_wait_en <= '0';

      CASE id_cs IS
         -- ----------------------------------------------------------------
         WHEN IDLE => -- 000
         -- ----------------------------------------------------------------
            dly_rst     <= '0';
            dly_ce      <= '0';
            dly_inc     <= '0';
            tap_wait_en <= '0';
            
            IF ( tap_cmd = TAP_CMD_RST ) THEN
               id_ns <= TAP_RST;
            ELSE
               IF ( tap_cmd = TAP_CMD_INC ) THEN
                  id_ns <= TAP_INC;
               ELSE
                  IF ( tap_cmd = TAP_CMD_DEC ) THEN
                     id_ns <= TAP_DEC;
                  ELSE
                     id_ns <= IDLE;
                  END IF;
               END IF;
            END IF;

         -- ----------------------------------------------------------------
         WHEN TAP_INC => -- 001
         -- ----------------------------------------------------------------
            dly_rst     <= '0';
            dly_ce      <= '1';
            dly_inc     <= '1';
            tap_wait_en <= '1';
            id_ns       <= TAP_WAIT;

         -- ----------------------------------------------------------------
         WHEN TAP_DEC => -- 002
         -- ----------------------------------------------------------------
            dly_rst     <= '0';
            dly_ce      <= '1';
            dly_inc     <= '0';
            tap_wait_en <= '1';
            id_ns       <= TAP_WAIT;

         -- ----------------------------------------------------------------
         WHEN TAP_RST => -- 004
         -- ----------------------------------------------------------------
            dly_rst     <= '1';
            dly_ce      <= '0';
            dly_inc     <= '0';
            tap_wait_en <= '1';
            id_ns       <= TAP_WAIT;

         -- ----------------------------------------------------------------
         WHEN TAP_WAIT => -- 008
         -- ----------------------------------------------------------------
            dly_rst     <= '0';
            dly_ce      <= '0';
            dly_inc     <= '0';
            tap_wait_en <= '0';

            IF ( tap_wait_cnt = "111" ) THEN
               id_ns <= IDLE;
            ELSE
               id_ns <= TAP_WAIT;
            END IF;

         -- ----------------------------------------------------------------
         WHEN OTHERS  =>
         -- ----------------------------------------------------------------
            id_ns <= IDLE;

      END CASE;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Wait counter used to allow IDELAY settling time 
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' OR NOT start_cal_r(5) = '1' ) THEN
            tap_wait_cnt <= "000";
         ELSE
            IF ( (tap_wait_cnt /= "111") OR tap_wait_en= '1' ) THEN
               tap_wait_cnt <= STD_LOGIC_VECTOR( UNSIGNED(tap_wait_cnt) + 1 );
            ELSE
               tap_wait_cnt <= tap_wait_cnt;
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- State machine "f_cs"
   --    Algorithm to find QK and DQ edges and align capture
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            f_cs <= Reset;   -- want to start out in RST
         ELSE
            f_cs <= f_ns;
         END IF;
      END IF;
   END PROCESS;

   -- combinatorial logic for "f_ns" state machine
   PROCESS (f_cs, cal_begin, tap_wait_cnt, first_edge_found, second_edge_found, 
            QK_tap_count_int, DQ_tap_count_int, DQ_bigger, DQ_centered_r, 
            QK_bigger, QK_centered_r, second_stage, tap_range, dq_error3, rst)
   BEGIN
      CASE f_cs IS
         -- ----------------------------------------------------------------
         WHEN Reset => -- 000
         -- ----------------------------------------------------------------
            IF ( cal_begin = '1' AND (tap_wait_cnt >= "110") ) THEN
               f_ns <= QK_INC;
            ELSE
               f_ns <= Reset;
            END IF;

         -- ----------------------------------------------------------------
         WHEN QK_INC => -- 001
         -- ----------------------------------------------------------------
         -- Increment QK to determine taps till edge of data valid window & save results
            IF ( first_edge_found = '1' AND second_edge_found = '1' ) THEN
               f_ns <= QK_RST;
            ELSE
               f_ns <= QK_INC;
            END IF;

         -- ----------------------------------------------------------------
         WHEN QK_RST => -- 002
         -- ----------------------------------------------------------------
         -- Decrement QK back down to 0
            -- Decrement Down to Zero instead of relying on RST
            IF ( QK_tap_count_int = "000000" ) THEN
               f_ns <= DQ_INC;
            ELSE
               f_ns <= QK_RST;
            END IF;

         -- ----------------------------------------------------------------
         WHEN DQ_INC => -- 004
         -- ----------------------------------------------------------------
         -- Increment DQ to determine taps till edge of data valid window & save results
            IF ( first_edge_found = '1' AND second_edge_found = '1' ) THEN
               f_ns <= DQ_RST;
            ELSE
               f_ns <= DQ_INC;
            END IF;

         -- ----------------------------------------------------------------
         WHEN DQ_RST => -- 008
         -- ----------------------------------------------------------------
         -- Decrement DQ back down to 0
            -- Decrement Down to Zero instead of relying on RST
            IF ( DQ_tap_count_int = "000000" ) THEN
               f_ns <= Center;
            ELSE
               f_ns <= DQ_RST;
            END IF;

         -- ----------------------------------------------------------------
         WHEN Center => -- 010
         -- ----------------------------------------------------------------
         -- Center QK in the DQ data valid window
            IF ( DQ_bigger = '1' AND NOT DQ_centered_r = '1' ) THEN   -- increment DQ
               f_ns <= Center;
            ELSE
               IF ( QK_bigger = '1' AND NOT QK_centered_r = '1' ) THEN   -- increment DQ
                  f_ns <= Center;
               ELSE
                  f_ns <= F_Done;   -- windows are the same, centered already
               END IF;
            END IF;

         -- ----------------------------------------------------------------
         WHEN F_Done => -- 020
         -- ----------------------------------------------------------------
         -- 1st stage calibration complete, wait for 2nd stage to begin
            IF ( second_stage = '1' ) THEN
               f_ns <= Both_INC;
            ELSE
               f_ns <= F_Done;
            END IF;

         -- ----------------------------------------------------------------
         WHEN Both_INC => -- 040
         -- ----------------------------------------------------------------
         -- Increment both QK/DQ to determine relation to FPGA clock (transfer clock domains)
            IF ( first_edge_found = '1' AND second_edge_found = '1' ) THEN
               f_ns <= Both_DEC;
            ELSE
               f_ns <= Both_INC;
            END IF;

         -- ----------------------------------------------------------------
         WHEN Both_DEC => -- 080
         -- ----------------------------------------------------------------
         -- Center QK/DQ together to transfer clock domains to FPGA clock
            IF ( QK_centered_r = '1' ) THEN
               f_ns <= S_Done;
            ELSE
               f_ns <= Both_DEC;
            END IF;

         -- ----------------------------------------------------------------
         WHEN DQ_DEC => -- 100
         -- ----------------------------------------------------------------
         -- Decrement DQ till 0 or error to move DQ around QK one last time
         -- Later this can be changed to Per-Bit Calibration if needed
            IF ( (DQ_tap_count_int = "000000") OR dq_error3 = '1' ) THEN
               f_ns <= DQ_INC2;
            ELSE
               f_ns <= DQ_DEC;
            END IF;

         -- ----------------------------------------------------------------
         WHEN DQ_INC2 => -- 200
         -- ----------------------------------------------------------------
         -- Increment DQ until window found
            IF ( first_edge_found = '1' AND second_edge_found = '1' AND (tap_range >= WINDOW_SIZE) ) THEN
               f_ns <= DQ_CENT;
            ELSE
               f_ns <= DQ_INC2;
            END IF;

         -- ----------------------------------------------------------------
         WHEN DQ_CENT => -- 400
         -- ----------------------------------------------------------------
         -- Center DQ in the data valid window
            IF ( DQ_centered_r = '1' ) THEN
               f_ns <= S_Done;
            ELSE
               f_ns <= DQ_CENT;
            END IF;

         -- ----------------------------------------------------------------
         WHEN S_Done => -- 800
         -- ----------------------------------------------------------------
         -- Calibration Done, wait here till reset.
            IF ( rst = '1' ) THEN
               f_ns <= Reset;
            ELSE
               f_ns <= S_Done;
            END IF;

         -- ----------------------------------------------------------------
         WHEN OTHERS  =>
         -- ----------------------------------------------------------------
            f_ns <= Reset;

      END CASE;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   --output generation
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         CASE f_cs IS
            -- ----------------------------------------------------------------
            WHEN Reset => -- 000
            -- ----------------------------------------------------------------
               first_cal_done_int <= '0';
               dly_cal_done_int   <= '0';
               rst_flags          <= '1';   -- reset the edge detection flags
               qk_flag            <= '1';
               dq_flag            <= '1';   -- determine who gets a signal (one or both)
               tap_cmd            <= TAP_CMD_RST;
               QK_tap_left        <= "000000";
               DQ_tap_left        <= "000000";
               QK_tap_window      <= "000000";
               DQ_tap_window      <= "000000";
               QK_tap_center      <= "000000";
               DQ_tap_center      <= "000000";
               first_edge_found   <= '0';
               second_edge_found  <= '0';
               tap_center         <= "000000";
               set_center         <= "000";

            -- ----------------------------------------------------------------
            WHEN QK_INC => -- 001
            -- ----------------------------------------------------------------
               first_cal_done_int <= '0';
               dly_cal_done_int   <= '0';
               rst_flags          <= '0';
               qk_flag            <= '1';
               dq_flag            <= '0';

               IF ( (QK_tap_count_int >= "001010") AND NOT first_edge_detect = '1' ) THEN
                  first_edge_found  <= '1';
                  second_edge_found <= '1';
               ELSE
                  IF ( first_edge_detect = '1' AND NOT first_edge_detect_r = '1' ) THEN
                     QK_tap_left      <= QK_tap_count_int;
                     first_edge_found <= '1';
                  ELSE
                     IF ( second_edge_detect = '1' AND NOT second_edge_detect_r = '1' ) THEN
                        QK_tap_window     <= tap_range;
                        second_edge_found <= '1';
                     END IF;
                  END IF;
               END IF;

               IF ( wait_for_refresh = '1' ) THEN
                  tap_cmd <= TAP_CMD_NOP;
               ELSE
                  tap_cmd <= TAP_CMD_INC;
               END IF;

            -- ----------------------------------------------------------------
            WHEN QK_RST => -- 002
            -- ----------------------------------------------------------------
               first_cal_done_int <= '0';
               dly_cal_done_int   <= '0';
               rst_flags          <= '1';
               qk_flag            <= '1';
               dq_flag            <= '0';
               first_edge_found   <= '0';
               second_edge_found  <= '0';
               tap_cmd            <= TAP_CMD_DEC;

            -- ----------------------------------------------------------------
            WHEN DQ_INC => -- 004
            -- ----------------------------------------------------------------
               first_cal_done_int <= '0';
               dly_cal_done_int   <= '0';
               rst_flags          <= '0';
               qk_flag            <= '0';
               dq_flag            <= '1';

               IF ( first_edge_detect = '1' AND NOT first_edge_detect_r = '1' ) THEN
                  DQ_tap_left      <= DQ_tap_count_int;
                  first_edge_found <= '1';
               ELSE
                  IF ( second_edge_detect = '1' AND NOT second_edge_detect_r = '1' ) THEN
                     DQ_tap_window <= tap_range;
                     tap_center <= STD_LOGIC_VECTOR( (UNSIGNED(tap_range) + UNSIGNED(QK_tap_window))/2 );
                     IF ( (tap_range < ('0' & WINDOW_SIZE(5 DOWNTO 1))) AND NOT end_of_taps = '1' ) THEN  -- compare with half of window size...
                        --check for at least small window
                        second_edge_found <= '0';
                        rst_flags         <= '1';
                     ELSE
                        second_edge_found <= '1';
                        rst_flags         <= '0';
                     END IF;
                  END IF;
               END IF;

               IF ( wait_for_refresh = '1' ) THEN
                  tap_cmd <= TAP_CMD_NOP;
               ELSE
                  tap_cmd <= TAP_CMD_INC;
               END IF;

            -- ----------------------------------------------------------------
            WHEN DQ_RST => -- 008
            -- ----------------------------------------------------------------
               first_cal_done_int <= '0';
               dly_cal_done_int   <= '0';
               rst_flags          <= '1';
               qk_flag            <= '0';
               dq_flag            <= '1';
               first_edge_found   <= '0';
               second_edge_found  <= '0';

               IF ( set_center /= "111" ) THEN
                  set_center <= STD_LOGIC_VECTOR( UNSIGNED(set_center) + 1 );
               END IF;

               IF ( set_center = "100" ) THEN
                  --DQ_tap_center <= DQ_tap_center_WORK;
                  --QK_tap_center <= QK_tap_center_WORK;
                  DQ_tap_center <= DQ_tap_center_WORK_r; --try registering these one more time for timing
                  QK_tap_center <= QK_tap_center_WORK_r;
               END IF;

               tap_cmd <= TAP_CMD_DEC;

            -- ----------------------------------------------------------------
            WHEN Center => -- 010
            -- ----------------------------------------------------------------
               first_cal_done_int <= '0';
               dly_cal_done_int   <= '0';
               rst_flags          <= '1';

               IF ( DQ_bigger_r = '1' ) THEN
                  -- increment DQ
                  qk_flag        <= '0';
                  dq_flag        <= '1';
                  tap_cmd        <= TAP_CMD_INC;
               ELSE
                  IF ( QK_bigger_r = '1' ) THEN
                     --increment QK
                     qk_flag        <= '1';
                     dq_flag        <= '0';
                     tap_cmd        <= TAP_CMD_INC;
                  ELSE
                     --windows are the same, centered already
                     qk_flag        <= '0';
                     dq_flag        <= '0';
                     tap_cmd        <= TAP_CMD_NOP;
                  END IF;
               END IF;

            -- ----------------------------------------------------------------
            WHEN F_Done => -- 020
            -- ----------------------------------------------------------------
            -- First stage of calibration is done, QK placed in center of DQ
               first_cal_done_int <= '1';
               dly_cal_done_int   <= '0';
               rst_flags          <= '1';
               qk_flag            <= '0';
               dq_flag            <= '0';
               --need to set this as a default value
               --need to ensure the algorithm doesn't think it is centered right away
               QK_tap_center <= "111111"; 
               tap_cmd            <= TAP_CMD_NOP;

            -- ----------------------------------------------------------------
            WHEN Both_INC => -- 040
            -- ----------------------------------------------------------------
               first_cal_done_int <= '1';
               dly_cal_done_int   <= '0';
               rst_flags          <= '0';
               qk_flag            <= '1';
               dq_flag            <= '1';

               IF ( first_edge_detect = '1' AND NOT first_edge_detect_r = '1' ) THEN
                  QK_tap_left      <= QK_tap_count_int;
                  DQ_tap_left      <= DQ_tap_count_int;
                  first_edge_found <= '1';
               ELSE
                  IF ( second_edge_detect = '1' AND NOT second_edge_detect_r = '1' ) THEN
                     second_edge_found <= '1';
                     IF ( inc_up2 = '0' ) THEN
                        QK_tap_center <= STD_LOGIC_VECTOR( UNSIGNED(QK_tap_count_int) - UNSIGNED(WINDOW_SIZE3) );     --decrement farther away from edge
                     ELSE
                        IF ( inc_up = '1') THEN
                           QK_tap_center <= STD_LOGIC_VECTOR( UNSIGNED(QK_tap_count_int) + UNSIGNED(WINDOW_SIZE2) );  --increment away from edge
                        ELSE
                           QK_tap_center <= STD_LOGIC_VECTOR( UNSIGNED(QK_tap_count_int) - UNSIGNED(WINDOW_SIZE2) );  --decrement away from edge
                        END IF;
                     END IF;
                  END IF;
               END IF;

               IF ( wait_for_refresh = '1' ) THEN
                  tap_cmd <= TAP_CMD_NOP;
               ELSE
                  tap_cmd <= TAP_CMD_INC;
               END IF;

            -- ----------------------------------------------------------------
            WHEN Both_DEC => -- 080
            -- ----------------------------------------------------------------
               first_cal_done_int <= '1';
               dly_cal_done_int   <= '0';
               rst_flags          <= '0';
               qk_flag            <= '1';
               dq_flag            <= '1';
               
               IF ( QK_tap_center < QK_tap_count_int ) THEN --want to inc or decrement away from edge
                  tap_cmd            <= TAP_CMD_DEC;
               ELSE
                  tap_cmd            <= TAP_CMD_INC;
               END IF;

            -- ----------------------------------------------------------------
            WHEN DQ_DEC => -- 100
            -- ----------------------------------------------------------------
               first_cal_done_int <= '1';
               dly_cal_done_int   <= '0';
               rst_flags          <= '1';
               qk_flag            <= '0';
               dq_flag            <= '1';
               first_edge_found   <= '0';
               second_edge_found  <= '0';

               IF ( wait_for_refresh = '1' ) THEN
                  tap_cmd <= TAP_CMD_NOP;
               ELSE
                  tap_cmd <= TAP_CMD_DEC;
               END IF;

            -- ----------------------------------------------------------------
            WHEN DQ_INC2 => -- 200
            -- ----------------------------------------------------------------
               first_cal_done_int <= '1';
               dly_cal_done_int   <= '0';
               rst_flags          <= '0';
               qk_flag            <= '0';
               dq_flag            <= '1';

               IF ( first_edge_detect = '1' AND NOT first_edge_detect_r = '1' ) THEN
                  DQ_tap_left      <= DQ_tap_count_int;
                  first_edge_found <= '1';
               ELSE
                  IF ( second_edge_detect = '1' AND NOT second_edge_detect_r = '1' ) THEN
                     DQ_tap_window     <= tap_range;
                     DQ_tap_center     <= STD_LOGIC_VECTOR( UNSIGNED(DQ_tap_left) + UNSIGNED(tap_range_center) );
                     second_edge_found <= '1';
                  END IF;
               END IF;

               IF ( second_edge_detect = '1' AND (tap_range < WINDOW_SIZE) ) THEN
                  rst_flags         <= '1';
                  first_edge_found  <= '0';
                  second_edge_found <= '0';
               END IF;

               IF ( wait_for_refresh = '1' ) THEN
                  tap_cmd <= TAP_CMD_NOP;
               ELSE
                  tap_cmd <= TAP_CMD_INC;
               END IF;

            -- ----------------------------------------------------------------
            WHEN DQ_CENT => -- 400
            -- ----------------------------------------------------------------
               first_cal_done_int <= '1';
               dly_cal_done_int   <= '0';
               rst_flags          <= '0';
               qk_flag            <= '0';
               dq_flag            <= '1';
               tap_cmd            <= TAP_CMD_DEC;

            -- ----------------------------------------------------------------
            WHEN S_Done => -- 800
            -- ----------------------------------------------------------------
               first_cal_done_int <= '1';
               dly_cal_done_int   <= '1';
               rst_flags          <= '0';
               qk_flag            <= '0';
               dq_flag            <= '0';
               tap_cmd            <= TAP_CMD_NOP;

            -- ----------------------------------------------------------------
            WHEN OTHERS  =>
            -- ----------------------------------------------------------------
               first_cal_done_int <= '0';
               dly_cal_done_int   <= '0';
               rst_flags          <= '1';   -- reset the edge detection flags
               qk_flag            <= '1';
               dq_flag            <= '1';   -- determine who gets a signal (one or both)
               tap_cmd            <= TAP_CMD_RST;
               QK_tap_left        <= "000000";
               DQ_tap_left        <= "000000";
               QK_tap_window      <= "000000";
               DQ_tap_window      <= "000000";
               QK_tap_center      <= "000000";
               DQ_tap_center      <= "000000";
               first_edge_found   <= '0';
               second_edge_found  <= '0';
               tap_center         <= "000000";
               set_center         <= "000";

         END CASE;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- First edge of the data valid window found
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' OR rst_flags_r = '1' ) THEN
            first_edge_detect <= '0';
         ELSE
            IF ( (NOT dq_error3 = '1' AND (tap_wait_cnt = "111") AND NOT wait_for_refresh = '1') OR end_of_taps = '1' ) THEN
               first_edge_detect <= '1';
            ELSE
               first_edge_detect <= first_edge_detect;
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Second Edge of the data valid window found
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' OR rst_flags_r = '1' ) THEN
            second_edge_detect <= '0';
         ELSE
            IF ( (dq_error3 = '1' AND first_edge_detect = '1' AND NOT wait_for_refresh = '1') OR end_of_taps = '1' ) THEN
               second_edge_detect <= '1';
            ELSE
               second_edge_detect <= second_edge_detect;
            END IF;
         END IF;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' OR rst_flags_r = '1' ) THEN
            first_edge_detect_r  <= '0';
            second_edge_detect_r <= '0';
         ELSE
            first_edge_detect_r  <= first_edge_detect;
            second_edge_detect_r <= second_edge_detect;
         END IF;
      END IF;
   END PROCESS;

   -- -----------------------------------------------------------------------------
   -- Separate tap counts for Strobe since is incremented seperatly from data
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( qk_dly_rst_int = '1' ) THEN
            QK_tap_count_int <= (others=>'0');
         ELSE
            IF ( qk_dly_ce_int = '1' AND qk_dly_inc_int = '1' ) THEN
               QK_tap_count_int <= STD_LOGIC_VECTOR( UNSIGNED(QK_tap_count_int) + 1 );
            ELSE
               IF ( qk_dly_ce_int = '1' AND NOT qk_dly_inc_int = '1' ) THEN
                  QK_tap_count_int <= STD_LOGIC_VECTOR( UNSIGNED(QK_tap_count_int) - 1 );
               ELSE
                  QK_tap_count_int <= QK_tap_count_int;
               END IF;
            END IF;
         END IF;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( dq_dly_rst_int = '1' ) THEN
            DQ_tap_count_int <= (others=>'0');
         ELSE
            IF ( dq_dly_ce_int = '1' AND dq_dly_inc_int = '1' ) THEN
               DQ_tap_count_int <= STD_LOGIC_VECTOR( UNSIGNED(DQ_tap_count_int) + 1 );
            ELSE
               IF ( dq_dly_ce_int = '1' AND NOT dq_dly_inc_int = '1' ) THEN
                  DQ_tap_count_int <= STD_LOGIC_VECTOR( UNSIGNED(DQ_tap_count_int) - 1 );
               ELSE
                  DQ_tap_count_int <= DQ_tap_count_int;
               END IF;
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- End of taps flag used to ensure we do not increment taps too much
   -- way to avoid tap jitter on data lines (clocks do not have pattern jitter)
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' OR rst_flags_r = '1' ) THEN
            end_of_taps <= '0';
         ELSE
            IF ( (DQ_tap_count_int >= DQ_TAP_END) OR (QK_tap_count_int >= QK_TAP_END) ) THEN
               end_of_taps <= '1';
            ELSE
               end_of_taps <= end_of_taps;
            END IF;
         END IF;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst_flags_r = '1' ) THEN
            tap_range <= (others=>'0');
         ELSE
            IF ( dly_inc = '1' AND first_edge_detect = '1' AND NOT second_edge_detect = '1' ) THEN
               tap_range <= STD_LOGIC_VECTOR( UNSIGNED(tap_range) + 1 );
            END IF;
         END IF;
      END IF;
   END PROCESS;

   tap_range_center <= '0' & tap_range(tap_range'high DOWNTO 1);  -- divide the count by 2, get the center value


   -- -----------------------------------------------------------------------------
   -- Want to Register Comparisons for better timing
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( DQ_tap_window > QK_tap_window ) THEN
            DQ_bigger <= '1';
            QK_bigger <= '0';
         ELSE
            IF ( DQ_tap_window < QK_tap_window ) THEN
               DQ_bigger <= '0';
               QK_bigger <= '1';
            ELSE
               DQ_bigger <= '0';
               QK_bigger <= '0';
            END IF;
         END IF;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         DQ_bigger_r <= DQ_bigger;
         QK_bigger_r <= QK_bigger;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( DQ_tap_count_int = DQ_tap_center ) THEN
            DQ_centered <= '1';
         ELSE
            DQ_centered <= '0';
         END IF;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( QK_tap_count_int = QK_tap_center ) THEN
            QK_centered <= '1';
         ELSE
            QK_centered <= '0';
         END IF;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         DQ_centered_r <= DQ_centered;
         QK_centered_r <= QK_centered;
         rst_flags_r   <= rst_flags;
      END IF;
   END PROCESS;

   --working registers perform temp calcs
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         --added
         DQ_tap_center_WORK_r <= DQ_tap_center_WORK;
         QK_tap_center_WORK_r <= QK_tap_center_WORK;
         IF ( set_center = "001" ) THEN
            DQ_tap_center_WORK <= STD_LOGIC_VECTOR( UNSIGNED(DQ_tap_window) - UNSIGNED(tap_center) );
            QK_tap_center_WORK <= STD_LOGIC_VECTOR( UNSIGNED(QK_tap_window) - UNSIGNED(tap_center) );
         ELSE
            IF ( set_center = "010" ) THEN
               DQ_tap_center_WORK <= STD_LOGIC_VECTOR( UNSIGNED(DQ_tap_center_WORK) + UNSIGNED(DQ_tap_left) );
               QK_tap_center_WORK <= STD_LOGIC_VECTOR( UNSIGNED(QK_tap_center_WORK) + UNSIGNED(QK_tap_left) );
            ELSE
               DQ_tap_center_WORK <= DQ_tap_center_WORK;
               QK_tap_center_WORK <= QK_tap_center_WORK;
            END IF;
         END IF;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE ( clk) THEN
         QK_tap_count_r <= QK_tap_count_int;
         DQ_tap_count_r <= DQ_tap_count_int;
      END IF;
   END PROCESS;

   --determine if we increment up or down when we find an edge for the 2nd stage transfer
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE ( clk) THEN
         IF ( QK_bigger_r = '1') THEN
            IF ( DQ_tap_count_r < STD_LOGIC_VECTOR( UNSIGNED(DQ_tap_left) + UNSIGNED(WINDOW_SIZE2) ) ) THEN
               inc_up <= '1';
            ELSE
               inc_up <= '0';
            END IF;
         ELSE
            IF ( QK_tap_count_r < STD_LOGIC_VECTOR( UNSIGNED(QK_tap_left) + UNSIGNED(WINDOW_SIZE2) ) ) THEN
               inc_up <= '1';
            ELSE
               inc_up <= '0';
            END IF;
         END IF;
      END IF;
   END PROCESS;
   
   --Want to check if we can decrement away even more than the minimum for better margin for 2nd stage
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE ( clk) THEN
         IF ( QK_bigger_r = '1') THEN
            IF ( DQ_tap_count_r < STD_LOGIC_VECTOR( UNSIGNED(DQ_tap_left) + UNSIGNED(WINDOW_SIZE3) ) ) THEN
               inc_up2 <= '1';
            ELSE
               inc_up2 <= '0';
            END IF;
         ELSE  --DQ has more taps then QK
            IF ( QK_tap_count_r < STD_LOGIC_VECTOR( UNSIGNED(QK_tap_left) + UNSIGNED(WINDOW_SIZE3) ) ) THEN
               inc_up2 <= '1';
            ELSE
               inc_up2 <= '0';
            END IF;
         END IF;
      END IF;
   END PROCESS;

END ARCHITECTURE rtl;
