--*****************************************************************************
-- DISCLAIMER OF LIABILITY
-- 
-- This text/file contains proprietary, confidential
-- information of Xilinx, Inc., is distributed under license
-- from Xilinx, Inc., and may be used, copied and/or
-- disclosed only pursuant to the terms of a valid license
-- agreement with Xilinx, Inc. Xilinx hereby grants you a 
-- license to use this text/file solely for design, simulation, 
-- implementation and creation of design files limited 
-- to Xilinx devices or technologies. Use with non-Xilinx 
-- devices or technologies is expressly prohibited and 
-- immediately terminates your license unless covered by
-- a separate agreement.
--
-- Xilinx is providing this design, code, or information 
-- "as-is" solely for use in developing programs and 
-- solutions for Xilinx devices, with no obligation on the 
-- part of Xilinx to provide support. By providing this design, 
-- code, or information as one possible implementation of 
-- this feature, application or standard, Xilinx is making no 
-- representation that this implementation is free from any 
-- claims of infringement. You are responsible for 
-- obtaining any rights you may require for your implementation. 
-- Xilinx expressly disclaims any warranty whatsoever with 
-- respect to the adequacy of the implementation, including 
-- but not limited to any warranties or representations that this
-- implementation is free from claims of infringement, implied 
-- warranties of merchantability or fitness for a particular 
-- purpose.
--
-- Xilinx products are not intended for use in life support
-- appliances, devices, or systems. Use in such applications is
-- expressly prohibited.
--
-- Any modifications that are made to the Source Code are 
-- done at the user's sole risk and will be unsupported.
--
-- Copyright (c) 2007 Xilinx, Inc. All rights reserved.
--
-- This copyright and support notice must be retained as part 
-- of this text at all times. 
--*****************************************************************************
--   ____  ____
--  /   /\/   /
-- /___/  \  /    Vendor: Xilinx
-- \   \   \/     Version: 1.1
--  \   \         Filename: rld_phy_calib.vhd
--  /   /         Timestamp: 24 August 2007
-- /___/   /\     
-- \   \  /  \
--  \___\/\___\
--
--
--	Device:  Virtex-4 and Virtex-5
--	Purpose: This module creates dummy read and write data for phy calibration  
-- 		   by writing data to FIFOs for processing by the controller.
--		      It supports BL=2 & BL=4.
-- Revision History:
--   Rev 1.0 - initial release
--   Rev 1.1 - Registered some of the FIFO control signals for better timing,
--             fixed issue with Config3/BL4 not producing back-to-back reads
--
--  Definition of "apAddr[25:0]" the user application address interface
--    [25]    [24]    [23]     [22:3]  [2:0]
--  user_ref, W-/R, reserved, address,  ba
--
--  bit[25]    user_ref  (0=disable user refresh i.e. automatic AREF; 1=enable user refresh)
--  bit[24]    W-/R      (0=write command; 1=read command)
--  bit[23]    reserved  (0)
--  bit[22:3]  address   (fixed 20-bit memory address)
--  bit[2:0]   ba        (3-bit memory bank address)
--
--*****************************************************************************

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;


ENTITY rld_phy_calib IS
   GENERIC (
      RL_DQ_WIDTH                    : integer := 36;    -- memory interface data width
      DEV_DQ_WIDTH                   : integer := 18;    -- data width of the memory device
      NUM_OF_DEVS                    : integer := 2;     --  number of memory devices
      NUM_OF_DKS                     : integer := 4; 
      DEV_AD_WIDTH                   : integer := 20;    -- address width of the memory device
      DEV_BA_WIDTH                   : integer := 3;     -- bank address width of the memory device
      APP_AD_WIDTH                   : integer := 26;
      DUPLICATE_CONTROLS             : boolean := FALSE; -- Duplicate the ports for A, BA, WE# and REF#
      BITS_PER_BYTE                  :  integer := 9;    -- Micron offers 9-bit byte for RL2
      RL_CK_PERIOD                   : integer := 3003;  -- CK clock period of the RLDRAM in ps
      -- MRS (Mode Register Set command) parameters   
      --    please check Micron RLDRAM-II datasheet for definitions of these parameters
      RL_MRS_CONF                    : std_logic_vector(2 DOWNTO 0) := "001";   -- "001": mode1;  "010": mode2;  "011": mode3
      RL_MRS_BURST_LENGTH            : std_logic_vector(1 DOWNTO 0) := "00";    -- "00": BL2;  "01": BL4;  "10": BL8
      RL_MRS_ADDR_MUX                : std_logic := '0';    -- '0': non-muxed addr;  '1': muxed addr
      RL_MRS_DLL_RESET               : std_logic := '0';    --
      RL_MRS_IMPEDANCE_MATCH         : std_logic := '1';    -- '0': internal 50ohms output buffer impedance, '1': external
      RL_MRS_ODT                     : std_logic := '0';    -- '0': disable term;  '1': enable term
      -- specific to FPGA/memory devices and capture method
      RL_IO_TYPE                     : std_logic_vector(1 DOWNTO 0) := "00";   -- CIO="00"  SIO_CIO_HYBRID="01"  SIO="10"
      DEVICE_ARCH                    : std_logic_vector(1 DOWNTO 0) := "00";   -- Virtex4="00"  Virtex5="01"
      CAPTURE_METHOD                 : std_logic_vector(1 DOWNTO 0) := "00";   -- Direct Clocking="00"  SerDes="01"
      -- specific to calibration scheme
      CAL_ADDRESS                    : std_logic_vector(19 DOWNTO 0) := (others => '0')  -- saved location to perform calibration
   );
   PORT (
      clk               : IN  std_logic;
      rstHard           : IN  std_logic;
      confBL            : IN  std_logic_vector(3 DOWNTO 0);  -- BL2=2, BL4=4, BL8=8 (in hex)
      start             : IN  std_logic;   
      first_sel_done    : IN  std_logic;  -- from phy, first calibration stage complete
      sel_done          : IN  std_logic;  -- calibration done
      refresh_cmd       : IN  std_logic;  -- for BL2 wait for refresh before aligning QVLD
      issueCalibration  : IN  std_logic;  
      rlWdfFull         : IN  std_logic;  -- write data fifo full
      rlWdfEmpty        : IN  std_logic;  -- write data fifo empty
      rlafFull          : IN  std_logic;  -- write address fifo full
      rlafAlmostEmpty   : IN  std_logic;  -- write address fifo almost full
      rlafEmpty         : IN  std_logic;  -- write address fifo empty
      second_cal_out    : OUT std_logic;  -- signal physical layer to begin 2nd phase cal
      qvld_cal_out      : OUT std_logic;  -- signal physical layer to align QVLD
      cal_done          : OUT std_logic;  -- calibration complete
      apValid           : OUT std_logic;  -- write enable to write address fifo
      apAddr            : OUT std_logic_vector(APP_AD_WIDTH-1 DOWNTO 0);     -- address to be written to write address fifo
      apWriteDValid     : OUT std_logic;  -- write enable to write data fifo
      apWriteData       : OUT std_logic_vector(2*RL_DQ_WIDTH-1 DOWNTO 0);  -- data to be written to write data fifo
      apWriteDM         : OUT std_logic_vector(2*NUM_OF_DEVS-1 DOWNTO 0);  -- data mask to be written to write data fifo
      cs_io             : INOUT std_logic_vector(1023 DOWNTO 0)   -- chipscope debug ports
   );
END ENTITY rld_phy_calib;



ARCHITECTURE rtl OF rld_phy_calib IS


   -- parameterize the pattern to be used for write data
   CONSTANT data1  : std_logic_vector(BITS_PER_BYTE-1 DOWNTO 0) := "111111111";   -- 9'h1FF  is  1_1111_1111
   CONSTANT data2  : std_logic_vector(BITS_PER_BYTE-1 DOWNTO 0) := "000000000";   -- 9'h000  is  0_0000_0000
   CONSTANT data3  : std_logic_vector(BITS_PER_BYTE-1 DOWNTO 0) := "010101010";   -- 9'h0AA  is  0_1010_1010
   CONSTANT data4  : std_logic_vector(BITS_PER_BYTE-1 DOWNTO 0) := "101010101";   -- 9'h155  is  1_0101_0101

   CONSTANT data_mask  : std_logic_vector(2*NUM_OF_DEVS-1 DOWNTO 0) := (others => '0'); 
   CONSTANT reserved   : std_logic := '0'; 
   CONSTANT user_ref   : std_logic := '0';


   SIGNAL count          : std_logic_vector(4 DOWNTO 0);
   SIGNAL count_2        : std_logic_vector(4 DOWNTO 0);
   SIGNAL qvld_wait_cnt  : std_logic_vector(4 DOWNTO 0);
   SIGNAL write_count    : std_logic_vector(3 DOWNTO 0);
   SIGNAL read_count     : std_logic_vector(2 DOWNTO 0);
   SIGNAL rlWdfEmpty_r   : std_logic;
   SIGNAL rlafEmpty_r    : std_logic;

   SIGNAL first_write_data   : std_logic_vector(2*RL_DQ_WIDTH-1 DOWNTO 0);
   SIGNAL second_write_data  : std_logic_vector(2*RL_DQ_WIDTH-1 DOWNTO 0);

   -- State Variables (using 1-hot encoding?)
   --    Reset   : std_logic_vector(6 DOWNTO 0) := "0000000";  -- not 1-hot
   --    Start   : std_logic_vector(6 DOWNTO 0) := "0000001";
   --    Write   : std_logic_vector(6 DOWNTO 0) := "0000010";
   --    Read    : std_logic_vector(6 DOWNTO 0) := "0000100";
   --    Wait    : std_logic_vector(6 DOWNTO 0) := "0001000";
   --    Rd_En   : std_logic_vector(6 DOWNTO 0) := "0010000";
   --    Rd_En2  : std_logic_vector(6 DOWNTO 0) := "0100000";
   --    Done    : std_logic_vector(6 DOWNTO 0) := "1000000";
--   type calibStates is ( Reset, Start, Write1, Write2, Read, Rd_en, Rd_en2, Done );
   type calibStates is ( Reset, Start_st, Write, Read, Wait_st, Rd_En, Rd_En2, Done );
   SIGNAL cs  : calibStates;  -- current state
   SIGNAL ns  : calibStates;  -- next state

   SIGNAL first_sel_done_r  : std_logic;
   SIGNAL second_cal        : std_logic;
   SIGNAL sel_done_delay    : std_logic;
   SIGNAL cal_done_r0       : std_logic;
   SIGNAL qvld_en           : std_logic;
   SIGNAL refresh_cmd_r     : std_logic;

   SIGNAL rst  : std_logic;
 
   SIGNAL burst_length          : std_logic_vector(3 DOWNTO 0);

   -- internal singals
   SIGNAL apValid_int           : std_logic;
   SIGNAL apAddr_int            : std_logic_vector(APP_AD_WIDTH-1 DOWNTO 0);
   SIGNAL apWriteDValid_int     : std_logic;
   SIGNAL apWriteData_int       : std_logic_vector(2*RL_DQ_WIDTH-1 DOWNTO 0);
   SIGNAL apWriteDM_int         : std_logic_vector(2*NUM_OF_DEVS-1 DOWNTO 0);


BEGIN
   cs_io          <= (others => 'Z');
   apValid        <= apValid_int;
   apAddr         <= apAddr_int;
   apWriteDValid  <= apWriteDValid_int;
   apWriteData    <= apWriteData_int;
   apWriteDM      <= apWriteDM_int;

   -- concatonate data to make up full datawidth
   write_data_18 : IF ( RL_DQ_WIDTH = 18) GENERATE
   BEGIN
      first_write_data  <= data1 & data1 & data2 & data2;
      second_write_data <= data3 & data3 & data4 & data4;
   END GENERATE write_data_18;
   
   write_data_36 : IF ( RL_DQ_WIDTH = 36) GENERATE
   BEGIN
      first_write_data  <= data1 & data1 & data1 & data1 & data2 & data2 & data2 & data2;
      second_write_data <= data3 & data3 & data3 & data3 & data4 & data4 & data4 & data4;
   END GENERATE write_data_36;
   
   write_data_72 : IF ( RL_DQ_WIDTH = 72) GENERATE
   BEGIN
      first_write_data  <= data1 & data1 & data1 & data1 & data1 & data1 & data1 & data1 &
                           data2 & data2 & data2 & data2 & data2 & data2 & data2 & data2;
      second_write_data <= data3 & data3 & data3 & data3 & data3 & data3 & data3 & data3 &
                           data4 & data4 & data4 & data4 & data4 & data4 & data4 & data4;
   END GENERATE write_data_72;
   
   rst          <= rstHard OR issueCalibration;
   burst_length <= confBL;

   -- -----------------------------------------------------------------------------
   -- Three stages required for Phy Calibration, keep the controller issuing back-to-back reads
   -- 1st Stage: align QK in DQ window for correct capture using Strobe
   -- 2nd Stage: align QK/DQ to FPGA clock domain for transfter
   -- 3rd Stage: align QVLD for all the bytes be performing one last read
   --    Note: BL2 needs two read commands that must be back-to-back
   --          wait for refresh then issue two read commands
   -- -----------------------------------------------------------------------------
   --update the current state
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN   -- sync reset
            
            cs <= Reset;
         ELSE
            cs <= ns;
         END IF;
      END IF;
   END PROCESS;

   -- next state generation
   PROCESS (cs, start, rlafEmpty, write_count, first_sel_done, first_sel_done_r, sel_done, burst_length, qvld_wait_cnt)
   BEGIN
      CASE cs IS
         WHEN Reset =>
                  ns <= Start_st;
         WHEN Start_st =>
                  IF ( (start AND rlafEmpty) = '1' ) THEN
                     ns <= Write;
                  ELSE
                     ns <= Start_st;
                  END IF;
         WHEN Write =>
                  -- write to FIFO, different banks to swamp bus with data
                  IF ( burst_length /= "0100" AND write_count = "0111" ) THEN
                     ns <= Read;
                  ELSE
                     IF ( burst_length = "0100" AND write_count = "1111" ) THEN
                        ns <= Read;
                     ELSE
                        ns <= Write;
                     END IF;
                  END IF;
         WHEN Read =>
                  -- wait in the read state until calibration finish
                  IF ( ((first_sel_done AND NOT first_sel_done_r) AND NOT sel_done) = '1' ) THEN
                     ns <= Start_st;
                  ELSE
                     IF ( (first_sel_done_r AND sel_done) = '1' ) THEN
                        IF ( burst_length = "0010" ) THEN
                           --only need to wait for BL2 operation
                           ns <= Wait_st;
                        ELSE
                           ns <= Rd_En;
                        END IF;
                     ELSE
                        ns <= Read;
                     END IF;
                  END IF;
         WHEN Wait_st =>
                  --for BL2 only, wait here till FIFO empty and refresh occurs
                  --need to ensure back-to-back data read for QVLD alignment
                  IF ( rlafEmpty = '1' AND qvld_wait_cnt = "11111" ) THEN
                     ns <= Rd_En;
                  ELSE
                     ns <= Wait_st;
                  END IF;
         WHEN Rd_En =>
                  -- Perform one last read to align QVLD
                  -- for BL2 need one extra read so go to Rd_En2 (BL4/8 only need 1 read command)
                  IF ( rlafEmpty = '1' ) THEN
                     IF ( burst_length = "0010" ) THEN
                        ns <= Rd_En2;
                     ELSE
                        ns <= Done;
                     END IF;
                  ELSE
                     ns <= Rd_En;
                  END IF;
         WHEN Rd_En2 =>
                  ns <= Done;
         WHEN Done =>
                  -- only a reset can force a state change
                  ns <= Done;
         WHEN OTHERS  =>
                  ns <= Reset;
      END CASE;
   END PROCESS;

   -- output generation
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         CASE cs IS
            -- -----------------------------------------------------------------------------
            WHEN Reset =>
            -- -----------------------------------------------------------------------------
               apWriteDValid_int   <= '0';
               apWriteData_int     <= (OTHERS => '0');
               apWriteDM_int       <= (OTHERS => '0');
               apValid_int         <= '0';
               apAddr_int          <= (OTHERS => '0');
               write_count         <= "0000";
               read_count          <= "000";
               second_cal          <= '0';
               sel_done_delay      <= '0';
               qvld_en             <= '0';

            -- -----------------------------------------------------------------------------
            WHEN Start_st =>
            -- -----------------------------------------------------------------------------
               apWriteDValid_int   <= '0';
               apWriteData_int     <= (OTHERS => '0');
               apWriteDM_int       <= (OTHERS => '0');
               apValid_int         <= '0';
               apAddr_int          <= (OTHERS => '0');
               write_count         <= "0000";
               read_count          <= "000";
               second_cal          <= '0';
               sel_done_delay      <= '0';
               qvld_en             <= '0';

            -- -----------------------------------------------------------------------------
            WHEN Write =>
            -- write to different bank location to ensure databus full of read data
            -- FIFO is filled with commands depending on burst length
            -- -----------------------------------------------------------------------------
               apWriteDValid_int   <= '1';
               apWriteDM_int       <= data_mask;
               IF ( first_sel_done = '1' AND first_sel_done_r = '1' ) THEN
                  second_cal <= '1';
               ELSE
                  second_cal <= '0';
               END IF;

               CASE write_count(2 DOWNTO 0) IS
                  -- -----------------------------
                  WHEN "000" =>
                  -- -----------------------------
                     apWriteData_int   <= first_write_data;
                     IF ( burst_length = "0010" ) THEN
                        apValid_int   <= '1';
                     ELSE
                        apValid_int   <= '0';
                     END IF;
                  -- -----------------------------
                  WHEN "001" =>
                  -- -----------------------------
                     IF ( first_sel_done = '1' ) THEN
                        apWriteData_int   <= second_write_data;
                     ELSE
                        apWriteData_int   <= first_write_data;
                     END IF;
                     --
                     IF ( burst_length = "0010" OR burst_length = "0100" ) THEN
                        apValid_int   <= '1';
                     ELSE
                        apValid_int   <= '0';
                     END IF;
                  -- -----------------------------
                  WHEN "010" =>
                  -- -----------------------------
                     apWriteData_int   <= first_write_data;
                     IF ( burst_length = "0010" ) THEN
                        apValid_int   <= '1';
                     ELSE
                        apValid_int   <= '0';
                     END IF;
                  -- -----------------------------
                  WHEN "011" =>
                  -- -----------------------------
                     IF ( first_sel_done = '1' ) THEN
                        apWriteData_int   <= second_write_data;
                     ELSE
                        apWriteData_int   <= first_write_data;
                     END IF;
                     --
                     apValid_int   <= '1';
                  -- -----------------------------
                  WHEN "100" =>
                  -- -----------------------------
                     apWriteData_int   <= first_write_data;
                     IF ( burst_length = "0010" ) THEN
                        apValid_int   <= '1';
                     ELSE
                        apValid_int   <= '0';
                     END IF;
                  -- -----------------------------
                  WHEN "101" =>
                  -- -----------------------------
                     IF ( first_sel_done = '1' ) THEN
                        apWriteData_int   <= second_write_data;
                     ELSE
                        apWriteData_int   <= first_write_data;
                     END IF;
                     --
                     IF ( burst_length = "0010" OR burst_length = "0100" ) THEN
                        apValid_int   <= '1';
                     ELSE
                        apValid_int   <= '0';
                     END IF;
                  -- -----------------------------
                  WHEN "110" =>
                  -- -----------------------------
                     apWriteData_int   <= first_write_data;
                     IF ( burst_length = "0010" ) THEN
                        apValid_int   <= '1';
                     ELSE
                        apValid_int   <= '0';
                     END IF;
                  -- -----------------------------
                  WHEN "111" =>
                  -- -----------------------------
                     IF ( first_sel_done = '1' ) THEN
                        apWriteData_int   <= second_write_data;
                     ELSE
                        apWriteData_int   <= first_write_data;
                     END IF;
                     --
                     apValid_int   <= '1';
                  -- -----------------------------
                  WHEN OTHERS  =>
                  -- -----------------------------
                     apWriteData_int   <= (OTHERS => '0');
                     apValid_int   <= '0';   
               END CASE;

               write_count <= STD_LOGIC_VECTOR( UNSIGNED(write_count) + 1 );

               --
               IF ( burst_length = "0010" ) THEN
                  apAddr_int   <= user_ref & '0' & reserved & CAL_ADDRESS(DEV_AD_WIDTH-1 DOWNTO 0) & write_count(2 DOWNTO 0);             -- write command (BL2)
               ELSE
                  IF ( burst_length = "0100" ) THEN
                     apAddr_int   <= user_ref & '0' & reserved & CAL_ADDRESS(DEV_AD_WIDTH-1 DOWNTO 0) & (write_count(3 DOWNTO 1));        -- write command (BL4)
                  ELSE
                     apAddr_int   <= user_ref & '0' & reserved & CAL_ADDRESS(DEV_AD_WIDTH-1 DOWNTO 0) & ('0' & write_count(3 DOWNTO 2));  -- write command (BL8)
                  END IF;
               END IF;

            -- -----------------------------------------------------------------------------
            WHEN Read =>
            -- read from different bank locations to ensure databus full of read data
            -- FIFO is filled with commands depending on burst length
            -- -----------------------------------------------------------------------------
               apWriteDValid_int   <= '0';
               apWriteData_int     <= (OTHERS => '0');
               apWriteDM_int       <= (OTHERS => '0');
               second_cal          <= second_cal;
               IF ( rlafAlmostEmpty = '1' ) THEN
                  read_count    <= STD_LOGIC_VECTOR( UNSIGNED(read_count) + 1 );
                  apValid_int   <= '1';
                  IF ( burst_length = "0010" OR burst_length = "0100") THEN
                     apAddr_int   <= user_ref & '1' & reserved & CAL_ADDRESS(DEV_AD_WIDTH-1 DOWNTO 0) & read_count;  -- read command (BL2 & BL4)
                  ELSE
                     apAddr_int   <= user_ref & '1' & reserved & CAL_ADDRESS(DEV_AD_WIDTH-1 DOWNTO 0) & (read_count AND "001");  -- read command (BL8)
                  END IF;
               ELSE
                  read_count    <= read_count;
                  apValid_int   <= '0';
                  apAddr_int    <= (OTHERS => '0');
               END IF;

            -- -----------------------------------------------------------------------------
            WHEN Wait_st =>
            -- -----------------------------------------------------------------------------
               IF ( (NOT refresh_cmd AND refresh_cmd_r) = '1' ) THEN
                  qvld_en <= '1';
               ELSE
                  qvld_en <= qvld_en;
               END IF;
               apWriteDValid_int   <= '0';
               apWriteData_int     <= (OTHERS => '0');
               apWriteDM_int       <= (OTHERS => '0');
               apValid_int         <= '0';
               apAddr_int          <= (OTHERS => '0');

            -- -----------------------------------------------------------------------------
            WHEN Rd_En =>
            -- -----------------------------------------------------------------------------
               -- Perform one last read for QVLD alignment
               IF ( rlafEmpty = '1' ) THEN
                  apValid_int   <= '1';
                  apAddr_int   <= user_ref & '1' & reserved & CAL_ADDRESS(DEV_AD_WIDTH-1 DOWNTO 0) & "000";  -- read command
               ELSE
                  apValid_int   <= '0';
                  apAddr_int    <= (OTHERS => '0');
               END IF;

            -- -----------------------------------------------------------------------------
            WHEN Rd_En2 =>
            -- -----------------------------------------------------------------------------
               -- this state only reached for BL2 to perform 1 extra read for QVLD alignment
               -- Bank address changed from previous state to ensure no gap in read
               apValid_int   <= '1';
               apAddr_int    <= user_ref & '1' & reserved & CAL_ADDRESS(DEV_AD_WIDTH-1 DOWNTO 0) & "001";

            -- -----------------------------------------------------------------------------
            WHEN Done =>
            -- -----------------------------------------------------------------------------
               apWriteDValid_int   <= '0';
               apWriteData_int     <= (OTHERS => '0');
               apWriteDM_int       <= (OTHERS => '0');
               apValid_int         <= '0';
               apAddr_int          <= (OTHERS => '0');
               sel_done_delay      <= '1';

            -- -----------------------------------------------------------------------------
            WHEN OTHERS  =>
            -- -----------------------------------------------------------------------------
               apWriteDValid_int   <= '0';
               apWriteData_int     <= (OTHERS => '0');
               apWriteDM_int       <= (OTHERS => '0');
               apValid_int         <= '0';
               apAddr_int          <= (OTHERS => '0');
               write_count         <= "0000";
               read_count          <= "000";
               qvld_en             <= '0';
         END CASE;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Generate delayed calibration_done signal to ensure calibration data is not written to FIFOs.
   -- Wait until calibration finishes and there is still data coming in
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            rlWdfEmpty_r <= '0';
            rlafEmpty_r  <= '0';
         ELSE
            rlWdfEmpty_r <= rlWdfEmpty;
            rlafEmpty_r  <= rlafEmpty;
         END IF;
      
      
         IF ( rst = '1' ) THEN
            count <= "00000";
         ELSE
            IF ( sel_done_delay = '1' AND rlWdfEmpty_r = '1' AND rlafEmpty_r = '1' AND (count /= "11111") ) THEN
               -- make sure all FIFOs empty before proceeding
               count <= STD_LOGIC_VECTOR( UNSIGNED(count) + 1 );
            ELSE
               count <= count;
            END IF;
         END IF;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            cal_done_r0 <= '0';
         ELSE
            IF ( count = "11111" ) THEN
               cal_done_r0 <= '1';
            END IF;
         END IF;
         IF ( rst = '1' ) THEN
            cal_done <= '0';
         ELSE
            cal_done <= cal_done_r0;  -- output delayed signal
         END IF;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            first_sel_done_r <= '0';
         ELSE
            first_sel_done_r <= first_sel_done;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------   
   -- Counter used to delay when calibration starts looking for 2nd stage data
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            count_2 <= "00000";
         ELSE
            IF ( second_cal = '1' AND (count_2 /= "11111") ) THEN
               count_2 <= STD_LOGIC_VECTOR( UNSIGNED(count_2) + 1 );
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------   
   -- output flag to indicate data belongs to 2nd part of calibration
   -- Phy needs to know when to start looking for 2nd stage data
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            second_cal_out <= '0';
         ELSE
            IF ( count_2 = "11111" ) THEN
               second_cal_out <= '1';
            ELSE
               second_cal_out <= '0';
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------   
   -- output flag to indicate ready for QVLD calibration (BL2 needs this
   -- to ensure the correct pattern is checked if refresh occurs).
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            qvld_wait_cnt <= "00000";
         ELSE
            IF ( qvld_en = '1' AND (qvld_wait_cnt /= "11111") ) THEN
               qvld_wait_cnt <= STD_LOGIC_VECTOR( UNSIGNED(qvld_wait_cnt) + 1 );
            ELSE
               qvld_wait_cnt <= qvld_wait_cnt;
            END IF;
         END IF;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            refresh_cmd_r <= '0';
         ELSE
            refresh_cmd_r <= refresh_cmd;
         END IF;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst = '1' ) THEN
            qvld_cal_out <= '0';
         ELSE
            qvld_cal_out <= sel_done_delay;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Simulation only, verification stage
   --    synthesis tool ignores assertions but simulator check them and report messages
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         assert ( (cs = Start_st) AND start = '1' AND NOT first_sel_done = '1' ) = FALSE  report " Starting Dummy Writes/Reads" severity NOTE;
         assert ( (cs = Start_st) AND start = '1' AND NOT first_sel_done = '1' ) = FALSE  report " First Stage Calibration Complete" severity NOTE;
         assert ( count = "11110" ) = FALSE  report "--------------- Calibration Done! --------------- Time: " severity NOTE;
      END IF;
   END PROCESS;


END ARCHITECTURE rtl;
