--*****************************************************************************
-- DISCLAIMER OF LIABILITY
-- 
-- This text/file contains proprietary, confidential
-- information of Xilinx, Inc., is distributed under license
-- from Xilinx, Inc., and may be used, copied and/or
-- disclosed only pursuant to the terms of a valid license
-- agreement with Xilinx, Inc. Xilinx hereby grants you a 
-- license to use this text/file solely for design, simulation, 
-- implementation and creation of design files limited 
-- to Xilinx devices or technologies. Use with non-Xilinx 
-- devices or technologies is expressly prohibited and 
-- immediately terminates your license unless covered by
-- a separate agreement.
--
-- Xilinx is providing this design, code, or information 
-- "as-is" solely for use in developing programs and 
-- solutions for Xilinx devices, with no obligation on the 
-- part of Xilinx to provide support. By providing this design, 
-- code, or information as one possible implementation of 
-- this feature, application or standard, Xilinx is making no 
-- representation that this implementation is free from any 
-- claims of infringement. You are responsible for 
-- obtaining any rights you may require for your implementation. 
-- Xilinx expressly disclaims any warranty whatsoever with 
-- respect to the adequacy of the implementation, including 
-- but not limited to any warranties or representations that this
-- implementation is free from claims of infringement, implied 
-- warranties of merchantability or fitness for a particular 
-- purpose.
--
-- Xilinx products are not intended for use in life support
-- appliances, devices, or systems. Use in such applications is
-- expressly prohibited.
--
-- Any modifications that are made to the Source Code are 
-- done at the user's sole risk and will be unsupported.
--
-- Copyright (c) 2007 Xilinx, Inc. All rights reserved.
--
-- This copyright and support notice must be retained as part 
-- of this text at all times. 
--*****************************************************************************
--   ____  ____
--  /   /\/   /
-- /___/  \  /    Vendor: Xilinx
-- \   \   \/     Version: 1.1
--  \   \         Filename: rld_rst.vhd
--  /   /         Timestamp: 24 August 2007
-- /___/   /\     
-- \   \  /  \
--  \___\/\___\
--
--
--	Device:  Virtex-4 and Virtex-5
--	Purpose: This module generates the reset signals and controls
--		the memory initial configuration (reset state machine).
-- Revision History:
--   Rev 1.0 - initial release
--   Rev 1.1 - Changed the initialization sequence to generate back-to-back
--             MRS commands per memory requirement along with holding the address
--             bus low during dummy MRS commands
--
--  Definition of "confMReg[17:0]" (MODE register)
--   [17:10]   [9]   [8]       [7]      [6]     [5]      [4:3]    [2:0]
--  RESERVED, TERM, RESIST, DLL_RESET,  '0', ADDR_MUX, BURST_LEN, CONF
--
--  bit[17:10] RESERVED
--  bit[9]     TERM      = '0'   (disable ODT)
--  bit[8]     RESIST    = '1'   (enable internal 50ohms output buffer impedance)
--  bit[7]     DLL_RESET = '0'   ()
--  bit[6]     ...reserved...
--  bit[5]     ADDR_MUX  = '0'   (non-multiplexed address)
--  bit[4:3]   BURST_LEN = "00"  ("00": BL2; "01": BL4; "10": BL8)
--  bit[2:0]   CONF      = "001" (config = 1, 2 or 3)
--
--*****************************************************************************

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;


ENTITY rld_rst IS
   GENERIC (
      SIMULATION_ONLY         : boolean := FALSE;  -- if TRUE, it shortens the wait time
      RL_DQ_WIDTH             : integer := 36;     -- memory interface data width
      DEV_DQ_WIDTH            : integer := 18;     -- data width of the memory device
      NUM_OF_DEVS             : integer := 2;      -- number of memory devices
      NUM_OF_DKS              : integer := 4;
      DEV_AD_WIDTH            : integer := 20;     -- address width of the memory device
      DEV_BA_WIDTH            : integer := 3;      -- bank address width of the memory device
      APP_AD_WIDTH            : integer := 26;
      DUPLICATE_CONTROLS      : boolean := FALSE;  -- Duplicate the ports for A, BA, WE# and REF#
      -- specific to FPGA/memory devices and capture method
      RL_IO_TYPE              : std_logic_vector(1 DOWNTO 0) := "00";  -- CIO="00"  SIO_CIO_HYBRID="01"  SIO="10"
      DEVICE_ARCH             : std_logic_vector(1 DOWNTO 0) := "01";  -- Virtex4="00"  Virtex5="01"
      CAPTURE_METHOD          : std_logic_vector(1 DOWNTO 0) := "01"   -- Direct Clocking="00"  SerDes="01"
   );
   PORT (
      clk                     : IN  std_logic;
      clk90                   : IN  std_logic;
      clk180                  : IN  std_logic;
      clk270                  : IN  std_logic;
      refClk                  : IN  std_logic;
      sysReset                : IN  std_logic;   -- system reset input, acive low, asynchronous

      confMReg                : IN  std_logic_vector(17 DOWNTO 0);
      dcmLocked               : IN  std_logic;
      issueMRS                : IN  std_logic;

      calibration_done        : IN  std_logic;   -- QK and DQ calibrations are completed

      sync_system_reset       : OUT std_logic;   -- system reset, synchronized to clk
      rstHard                 : OUT std_logic;
      rstHard_90              : OUT std_logic;
      rstHard_180             : OUT std_logic;
      rstHard_270             : OUT std_logic;
      rstHard_refClk          : OUT std_logic;
      rstConfig               : OUT std_logic;

      Init_Done               : OUT std_logic;
      Init_Done_270           : OUT std_logic;
      mrsDone_o               : OUT std_logic;

      rstCmd                  : OUT std_logic_vector(3 DOWNTO 0);
      rstAd                   : OUT std_logic_vector(DEV_AD_WIDTH-1 DOWNTO 0);
      rstBa                   : OUT std_logic_vector(DEV_BA_WIDTH-1 DOWNTO 0)
   );
END ENTITY rld_rst;



ARCHITECTURE rtl OF rld_rst IS


   -- simulation only setting to shorten the wait time (when SIMULATION_ONLY is TRUE)
   CONSTANT REF_VAL_shorten                    : std_logic_vector(10 DOWNTO 0) := "00000110010";   -- X"032", refresh counts
   CONSTANT REF_TRC_shorten                    : std_logic_vector(10 DOWNTO 0) := "00000110010";   -- X"032"
   CONSTANT WAIT_FOR_POWER_STABLE_CNT_shorten  : std_logic_vector(15 DOWNTO 0) := "0000000000110010";   -- X"0032", 50 ticks @210MHz is 0.238us
   -- normal settings for wait time (when SIMULATION_ONLY is FALSE)
   CONSTANT REF_VAL                    : std_logic_vector(10 DOWNTO 0) := "11111111111";   -- X"7FF", refresh counts
   CONSTANT REF_TRC                    : std_logic_vector(10 DOWNTO 0) := "11111101100";   -- X"7EC"
   CONSTANT WAIT_FOR_POWER_STABLE_CNT  : std_logic_vector(15 DOWNTO 0) := "1010010000010000";   -- X"A410", 420000 ticks @210MHz is 200us
   --
   --  defines "rstSt" possible states
--   CONSTANT WAIT200us        : std_logic_vector(2 DOWNTO 0) := "000";
--   CONSTANT RESET            : std_logic_vector(2 DOWNTO 0) := "001";
--   CONSTANT CONFIG           : std_logic_vector(2 DOWNTO 0) := "010";
--   CONSTANT REFRESH          : std_logic_vector(2 DOWNTO 0) := "011";
--   CONSTANT ALL_DONE         : std_logic_vector(2 DOWNTO 0) := "100";
--   CONSTANT MODE_CHANGE      : std_logic_vector(2 DOWNTO 0) := "101";
--   CONSTANT WAIT_DLL_STABLE  : std_logic_vector(2 DOWNTO 0) := "110";
   type rstStStates is ( WAIT200us, RESET, CONFIG, REFRESH, ALL_DONE, MODE_CHANGE, WAIT_DLL_STABLE );
   SIGNAL rstSt  : rstStStates;

   --
   --  defines "rstCmd" possible commands
   --                                                 CS_N  AS_N  WE_N  REF_N
   CONSTANT MRS_CMD  : std_logic_vector(3 DOWNTO 0) := '0' & '0' & '0' & '0'; 
   CONSTANT NOP_CMD  : std_logic_vector(3 DOWNTO 0) := '1' & '1' & '1' & '1'; 
   CONSTANT REF_CMD  : std_logic_vector(3 DOWNTO 0) := '0' & '1' & '1' & '0';


   SIGNAL hardInt                     : std_logic;

   SIGNAL bank                        : std_logic_vector(3-1 DOWNTO 0);
   SIGNAL refCnt                      : std_logic_vector(10 DOWNTO 0);
   SIGNAL WaitCnt                     : std_logic_vector(15 DOWNTO 0);
   SIGNAL cfgCnt                      : std_logic_vector(4 DOWNTO 0);
   SIGNAL mrsCnt                      : std_logic_vector(2 DOWNTO 0);
   SIGNAL waitTrc                     : std_logic;
   SIGNAL doRefresh                   : std_logic;
   SIGNAL refStart                    : std_logic;

   SIGNAL rstHard_0                   : std_logic;
   SIGNAL rstHard_a1_0                : std_logic;
   SIGNAL rstHard_a1_180              : std_logic;
   SIGNAL rstHard_a1_270              : std_logic;
   SIGNAL rstHard_a2_270              : std_logic;

   SIGNAL Init_Done_a1_270            : std_logic;

   SIGNAL rstHard_a1_refClk           : std_logic;
   SIGNAL rstHard_a2_refClk           : std_logic;

   SIGNAL Detect_final_WaitCnt_refClk : std_logic;   -- on refClk clock domain
   SIGNAL Detect_final_WaitCnt        : std_logic;   -- on clk clock domain
   SIGNAL Detect_final_WaitCnt_d1     : std_logic;   -- on clk clock domain

   SIGNAL addr_mux                    : std_logic;

   SIGNAL mr_dll_reset_bit            : std_logic;
   SIGNAL prev_mr_dll_reset_bit       : std_logic;
   SIGNAL dll_reset_change            : std_logic;
   SIGNAL dll_stable_wait_cnt         : std_logic_vector(9 DOWNTO 0);
   SIGNAL dll_is_stable               : std_logic;

   SIGNAL sys_rst_0                   : std_logic;
   SIGNAL sys_rst_1                   : std_logic;
   SIGNAL sys_rst_2                   : std_logic;
   SIGNAL sys_rst_3                   : std_logic;
   SIGNAL sync_system_reset_pre       : std_logic;
   SIGNAL sync_system_reset_int       : std_logic;   -- internal version of "sync_system_reset"

   SIGNAL vcc                         : std_logic;
   SIGNAL gnd                         : std_logic;

   -- internal signals
   SIGNAL rstHard_refClk_int          : std_logic;
   SIGNAL Init_Done_int               : std_logic;
   SIGNAL rstAd_int                   : std_logic_vector(DEV_AD_WIDTH-1 DOWNTO 0);


BEGIN
   vcc                    <= '1';
   gnd                    <= '0';

   rstHard_refClk         <= rstHard_refClk_int;
   Init_Done              <= Init_Done_int;
   rstAd                  <= rstAd_int;

   rstBa                  <= bank;

   sync_system_reset      <= sync_system_reset_int;  -- internal copy of sync_system_reset

   rstHard                <= rstHard_0;  -- align 0-deg reset with the others

   addr_mux               <= confMReg(5);   -- bit for ADDR_MUX, multiplexed address
   mr_dll_reset_bit       <= confMReg(7);   -- bit for DLL_RESET
   prev_mr_dll_reset_bit  <= rstAd_int(7);  -- bit for DLL_RESET, previously set


   -- -----------------------------------------------------------------------------
   -- Generate a stable synchronous reset for the memory interface
   --   - asynchronously asserted, with system reset and DCM Lock
   --   - synchronously deasserted, once both sources of reset are gone
   -- 
   -- (1) sysReset is asynchronous, coming from an external source
   -- (2) The DCM lock signal is considered asynchronous to the clock
   -- 
   -- Once sysReset is disabled (active high), the DCM starts acquiring lock;
   -- this will take some time.
   -- While dcmLocked stays low, NOT dcmLocked is high and forces reset.
   -- Once the DCM reaches "LOCKED", it will disable the asynchronous reset.
   -- -----------------------------------------------------------------------------
   sync_system_reset_pre <= (NOT sysReset) OR (NOT dcmLocked);

   PROCESS (clk, sync_system_reset_pre)   -- it uses an asynchronous reset...
   BEGIN
      IF (sync_system_reset_pre = '1') THEN
         sys_rst_0 <= '1';   -- enable sync reset
         sys_rst_1 <= '1';
         sys_rst_2 <= '1';
         sys_rst_3 <= '1';
         sync_system_reset_int <= '1';
      ELSIF RISING_EDGE( clk ) THEN
         sys_rst_0 <= '0';        -- disable reset, inject zero
         sys_rst_1 <= sys_rst_0;  -- propagate the zero
         sys_rst_2 <= sys_rst_1;
         sys_rst_3 <= sys_rst_2;
         sync_system_reset_int <= sys_rst_3;   -- on the fifth clock, reset is disabled
      END IF;
   END PROCESS;

   -- -----------------------------------------------------------------------------
   -- Sequence of reset for different clock domains, related to system clock
   --   - ripple main reset signal (hardInt) through 3/4 cycle minimum
   --   - all reset attached to system clock are synchronous
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         rstHard_a1_0 <= hardInt;       -- capture main reset with full cycle
         rstHard_0    <= rstHard_a1_0;  -- create rstHard 0-deg reset (full cycle)
      END IF;
   END PROCESS;

   PROCESS (clk90)
   BEGIN
      IF RISING_EDGE( clk90 ) THEN
         rstHard_90 <= rstHard_a1_180;  -- capture 180-deg reset with 3/4 cycle
      END IF;
   END PROCESS;

   PROCESS (clk180)
   BEGIN
      IF RISING_EDGE( clk180 ) THEN
         rstHard_a1_180 <= rstHard_a1_270;  -- capture 270-deg reset with 3/4 cycle
         rstHard_180    <= rstHard_a1_180;  -- create rstHard 180-deg reset (full cycle)
      END IF;
   END PROCESS;

   PROCESS (clk270)
   BEGIN
      IF RISING_EDGE( clk270 ) THEN
         rstHard_a1_270 <= hardInt;         -- capture main reset with 3/4 cycle
         rstHard_a2_270 <= rstHard_a1_270;  -- re-align reset (full cycle)
         rstHard_270    <= rstHard_a2_270;  -- create rstHard 270-deg reset (full cycle)
      END IF;
   END PROCESS;

   -- -----------------------------------------------------------------------------
   -- Reset sequence related to refClk clock
   --   - reset is controlled by FSM and it is a very long period (200us) => hardInt
   --   - capture main reset signal (hardInt) through metastable flops
   --   - system clock can vary from 175MHz (min memory's DLL) to 500MHz (max DLL)
   --   - so a clock period range of 5.715ns (~175MHz) down to 2.000ns (500MHz)
   --   - need one to three cycles delay which depends on system clock
   --   - delay has no impact on design, system clock module waits for refClk ready
   -- -----------------------------------------------------------------------------
   PROCESS (refClk)
   BEGIN
      IF RISING_EDGE( refClk ) THEN
         rstHard_a1_refClk  <= sync_system_reset_int;  -- capture sync system reset (unknown phase but stable)
         rstHard_a2_refClk  <= rstHard_a1_refClk;      -- remove possible metastability (flops sequence)
         rstHard_refClk_int <= rstHard_a2_refClk;      -- create rstHard reset (full refClk cycle)
      END IF;
   END PROCESS;

   -- -----------------------------------------------------------------------------
   -- Init_Done, related to 270-deg system clock
   -- -----------------------------------------------------------------------------
   PROCESS (clk270)
   BEGIN
      IF RISING_EDGE( clk270 ) THEN
         Init_Done_a1_270 <= Init_Done_int;     -- capture Init_Done with 3/4 cycle
         Init_Done_270    <= Init_Done_a1_270;  -- create Init_Done 270-deg (full cycle)
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Reset state machine
   --   - using stable sync system reset "sync_system_reset_int"
   -- -----------------------------------------------------------------------------
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (sync_system_reset_int = '1') THEN
            rstSt <= WAIT200us;
         ELSE
            IF (rstSt = WAIT200us AND Detect_final_WaitCnt = '1') THEN
               rstSt <= RESET;
            ELSE
               IF (rstSt = RESET) THEN
                  rstSt <= CONFIG;
               ELSE
                  IF (rstSt = CONFIG AND cfgCnt = "00000") THEN
                     rstSt <= REFRESH;
                  ELSE
                     IF (rstSt = REFRESH AND waitTrc = '1' AND bank = "110") THEN
                        rstSt <= WAIT_DLL_STABLE;
                     ELSE
                        IF (rstSt = WAIT_DLL_STABLE AND dll_is_stable = '1') THEN
                           rstSt <= ALL_DONE;
                        ELSE
                           IF (rstSt = ALL_DONE AND issueMRS = '1') THEN
                              rstSt <= MODE_CHANGE;
                           ELSE
                              IF ( rstSt = MODE_CHANGE AND mrsCnt = "000" ) THEN
                                 IF ( dll_reset_change = '1' ) THEN
                                    rstSt <= WAIT_DLL_STABLE;
                                 ELSE
                                    rstSt <= ALL_DONE;
                                 END IF;
                              END IF;
                           END IF;
                        END IF;
                     END IF;
                  END IF;
               END IF;
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   -- Wait for some time before releasing rstConfig after completing the 
   -- REFRESH states specs ask for wait of Trc.  We wait for more time than that.
   -- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ((SIMULATION_ONLY = FALSE AND refCnt = REF_TRC) OR (SIMULATION_ONLY = TRUE AND refCnt = REF_TRC_shorten)) THEN
            waitTrc <= '1';
         ELSE
            waitTrc <= '0';
         END IF;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (rstSt = RESET) THEN
            doRefresh <= '0';
         ELSE
            IF ((SIMULATION_ONLY = FALSE AND refCnt = REF_VAL) OR (SIMULATION_ONLY = TRUE AND refCnt = REF_VAL_shorten)) THEN
               doRefresh <= '1';
            ELSE
               doRefresh <= '0';
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   -- Generation of external signals
   -- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (sync_system_reset_int = '1') THEN
            hardInt       <= '1';
            rstConfig     <= '0';
            Init_Done_int <= '0';
            rstCmd        <= NOP_CMD;
            rstAd_int     <= (OTHERS => '0');
            bank          <= (OTHERS => '0');
            refStart      <= '1';
         ELSE
            IF (rstSt = RESET) THEN
               hardInt       <= '0';
               rstConfig     <= '1';
               Init_Done_int <= '0';
               rstCmd        <= NOP_CMD;
               rstAd_int     <= (OTHERS => '0');   -- keep the address bus low
               bank          <= (OTHERS => '1');
            ELSE
               IF (rstSt = CONFIG) THEN
                  CASE cfgCnt IS
                     WHEN "01110" => rstCmd     <= MRS_CMD;
                                     rstAd_int  <= (OTHERS => '0');
                     WHEN "01101" => rstCmd     <= MRS_CMD;
                     WHEN "01100" => rstCmd     <= MRS_CMD;
                                     rstAd_int  <= '0' & '0' & confMReg(17 DOWNTO 0);
                     WHEN OTHERS  => rstCmd     <= NOP_CMD;
                  END CASE;
               ELSE
                  IF (rstSt = REFRESH AND doRefresh = '1') THEN
                     rstCmd <= REF_CMD;
                     IF (refStart = '1') THEN
                        refStart <= '0';
                     ELSE
                        CASE bank IS
                           WHEN "111"  => bank <= (OTHERS => '0');
                           WHEN "000"  => bank <= "001";
                           WHEN "001"  => bank <= "010";
                           WHEN "010"  => bank <= "011";
                           WHEN "011"  => bank <= "100";
                           WHEN "100"  => bank <= "101";
                           WHEN "101"  => bank <= "110";
                           WHEN "110"  => bank <= (OTHERS => '1');
                           WHEN OTHERS => NULL;
                        END CASE;
                     END IF;
                  ELSE
                     IF (rstSt = REFRESH) THEN
                        rstCmd <= NOP_CMD;
                     ELSE
                        IF (rstSt = WAIT_DLL_STABLE) THEN
                           rstConfig <= '0';
                        ELSE
                           IF (rstSt = ALL_DONE) THEN
                              rstConfig <= '0';
                              -- make sure  "calibration_done = 1", i.e. tap calibration is complete
                              Init_Done_int <= calibration_done;
                           ELSE
                              IF (rstSt = MODE_CHANGE) THEN
                                 rstConfig <= '1';
                                 CASE mrsCnt IS
                                    WHEN "110"  => rstCmd <= MRS_CMD;
                                    WHEN OTHERS => rstCmd <= NOP_CMD;
                                 END CASE;
                                 rstAd_int <= '0' & '0' & confMReg;
                                 bank <= (OTHERS => '1');
                              END IF;
                           END IF;
                        END IF;
                     END IF;
                  END IF;
               END IF;
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   -- Refresh counter
   -- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (sync_system_reset_int = '1' OR rstSt = RESET OR (refCnt = "00000000000")) THEN
            IF (SIMULATION_ONLY = FALSE) THEN
               refCnt <= REF_VAL;
            ELSE
               -- minimize the wait time when SIMULATION_ONLY is set
               refCnt <= REF_VAL_shorten;
            END IF;
         ELSE
            IF (rstSt = CONFIG) THEN
               refCnt <= STD_LOGIC_VECTOR( UNSIGNED(refCnt) - 1 );   -- decrement
            ELSE
               IF (rstSt = REFRESH) THEN
                  refCnt <= STD_LOGIC_VECTOR( UNSIGNED(refCnt) - 1 );   -- decrement
               END IF;
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   -- Config counter
   -- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~   
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (sync_system_reset_int = '1' OR rstSt = RESET) THEN
            IF (addr_mux = '1') THEN
               cfgCnt <= "10101";   -- 1_0101
            ELSE
               cfgCnt <= "01111";   -- 0_1111
            END IF;
         ELSE
            IF (rstSt = CONFIG) THEN
               cfgCnt <= STD_LOGIC_VECTOR( UNSIGNED(cfgCnt) - 1 );   -- decrement
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   -- MRS counter to handle Trc
   -- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~   
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (sync_system_reset_int = '1' OR rstSt = ALL_DONE) THEN
            mrsCnt <= "111";
         ELSE
            IF (rstSt = MODE_CHANGE) THEN
               mrsCnt <= STD_LOGIC_VECTOR( UNSIGNED(mrsCnt) - 1 );   -- decrement
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- -------------------------------------------------------------------------
   -- Indicates MRS is done
   -- -------------------------------------------------------------------------   
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (sync_system_reset_int = '1' OR issueMRS = '1') THEN
            mrsDone_o <= '0';
         ELSE
            IF (rstSt = ALL_DONE) THEN
               mrsDone_o <= '1';
            END IF;
         END IF;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (sync_system_reset_int = '1') THEN
            dll_reset_change <= '0';
         ELSE
            IF ((rstSt = MODE_CHANGE) AND (mrsCnt = "111")) THEN
               dll_reset_change <= prev_mr_dll_reset_bit XOR mr_dll_reset_bit;  -- true when not equal
            END IF;
         END IF;
      END IF;
   END PROCESS;

   -- counter to wait for dll is stable (per Micron RLDRAM II Design Guide),
   -- need to wait 1024 cycles for the dll locks
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (sync_system_reset_int = '1' OR issueMRS = '1') THEN
            dll_stable_wait_cnt <= (others => '0');
         ELSE
            IF (rstSt = WAIT_DLL_STABLE) THEN
               dll_stable_wait_cnt <= STD_LOGIC_VECTOR( UNSIGNED(dll_stable_wait_cnt) + 1 );
            END IF;
         END IF;
      END IF;
   END PROCESS;

   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF (sync_system_reset_int = '1' OR issueMRS = '1') THEN
            dll_is_stable <= '0';
         ELSE
            IF (dll_stable_wait_cnt = "1111111111") THEN
               dll_is_stable <= '1';
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- -----------------------------------------------------------------------------
   -- Wait for a minimum of 200 us for memory initialization 
   -- Use fixed reference clock refClk 200MHz for this counter
   --
   -- Virtex-4 and Virtex-5 min-typ-max tolerance for refClk (200MHz +/- 10MHz)
   --                                  
   -- 		freq    count   wait     wait time
   --                  (ticks)  time     (use fixed "high" count)
   -- high      210MHz  42000   200us    200us
   -- typical   200MHz  40000   200us    210us
   -- low       190MHz  38000   200us    221.026316us
   --
   -- Counter must used max frequency tolerance of refClk 210MHz (200MHz + 10MHz)
   -- to cover the lowest clock period for the counter and guarantee a minimum of
   -- 200us in all tolerances.
   -- 
   -- Max count at 210MHz is 42000 ticks, it requires 16-bit counter (~15.358 bits).
   -- A fixed counter at 210MHz does not require a pre-scaler or a divide-by-x clock.
   -- Of course, the count value will not depend on the input frequency.
   --
   -- WaitCnt is compared against the parameter WAIT_FOR_POWER_STABLE_CNT.
   --
   -- Add simulation-only parameter that minimizes wait time (SIMULATION_ONLY).
   -- -----------------------------------------------------------------------------
   PROCESS (refClk)
   BEGIN
      IF RISING_EDGE( refClk ) THEN
         IF (rstHard_refClk_int = '1') THEN
            WaitCnt <= (others => '0');
         ELSE
            IF (Detect_final_WaitCnt_refClk = '0') THEN
               WaitCnt <= STD_LOGIC_VECTOR( UNSIGNED(WaitCnt) + 1 );   -- increment
            END IF;
         END IF;
      END IF;
   END PROCESS;

   simulation_only_disabled : PROCESS (refClk)
   BEGIN
      IF RISING_EDGE( refClk ) THEN
         IF (rstHard_refClk_int = '1') THEN
            Detect_final_WaitCnt_refClk <= '0';
         ELSE
            IF ((SIMULATION_ONLY = FALSE AND WaitCnt = WAIT_FOR_POWER_STABLE_CNT) OR (SIMULATION_ONLY = TRUE AND WaitCnt = WAIT_FOR_POWER_STABLE_CNT_shorten)) THEN
               Detect_final_WaitCnt_refClk <= '1';   -- stays set once it sees "max count"
            END IF;
         END IF;
      END IF;
   END PROCESS simulation_only_disabled;

   -- Transfer to system clock domain, metastable flops in place
   --   - no reset needed, Detect_final_WaitCnt_refClk is stable
   --     long before the state machine rstSt comes out of reset.
   --   - once set, Detect_final_WaitCnt_refClk is stable until 
   --     another system reset occurs
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         Detect_final_WaitCnt_d1 <= Detect_final_WaitCnt_refClk;   -- capture signal in clk domain
         Detect_final_WaitCnt    <= Detect_final_WaitCnt_d1;
      END IF;
   END PROCESS;


END ARCHITECTURE rtl;
