--*****************************************************************************
-- DISCLAIMER OF LIABILITY
-- 
-- This text/file contains proprietary, confidential
-- information of Xilinx, Inc., is distributed under license
-- from Xilinx, Inc., and may be used, copied and/or
-- disclosed only pursuant to the terms of a valid license
-- agreement with Xilinx, Inc. Xilinx hereby grants you a 
-- license to use this text/file solely for design, simulation, 
-- implementation and creation of design files limited 
-- to Xilinx devices or technologies. Use with non-Xilinx 
-- devices or technologies is expressly prohibited and 
-- immediately terminates your license unless covered by
-- a separate agreement.
--
-- Xilinx is providing this design, code, or information 
-- "as-is" solely for use in developing programs and 
-- solutions for Xilinx devices, with no obligation on the 
-- part of Xilinx to provide support. By providing this design, 
-- code, or information as one possible implementation of 
-- this feature, application or standard, Xilinx is making no 
-- representation that this implementation is free from any 
-- claims of infringement. You are responsible for 
-- obtaining any rights you may require for your implementation. 
-- Xilinx expressly disclaims any warranty whatsoever with 
-- respect to the adequacy of the implementation, including 
-- but not limited to any warranties or representations that this
-- implementation is free from claims of infringement, implied 
-- warranties of merchantability or fitness for a particular 
-- purpose.
--
-- Xilinx products are not intended for use in life support
-- appliances, devices, or systems. Use in such applications is
-- expressly prohibited.
--
-- Any modifications that are made to the Source Code are 
-- done at the user's sole risk and will be unsupported.
--
-- Copyright (c) 2007 Xilinx, Inc. All rights reserved.
--
-- This copyright and support notice must be retained as part 
-- of this text at all times. 
--*****************************************************************************
--   ____  ____
--  /   /\/   /
-- /___/  \  /    Vendor: Xilinx
-- \   \   \/     Version: 1.0
--  \   \         Filename: rld_seven_seg.vhd
--  /   /         Timestamp: 24 August 2007
-- /___/   /\     
-- \   \  /  \
--  \___\/\___\
--
--
--	Device:  Virtex-5
--	Purpose: This module generates external signals for LEDs and the 7-segment
--		display on the ML561.  Can be used for Demo or Debug.
--
--*****************************************************************************

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_std.ALL;


ENTITY rld_seven_seg IS
   GENERIC (
      READ_DIV_CNT_WIDTH   : integer := 28
   );
   PORT (
      clk                     : IN  std_logic;
      rst_n                   : IN  std_logic;

      init_done               : IN  std_logic;
      valid                   : IN  std_logic;
      PASS_FAIL               : IN  std_logic_vector(2 DOWNTO 0);

      seven_seg_n             : OUT std_logic_vector(6 DOWNTO 0);
      seven_seg_dp_n          : OUT std_logic;
      LED                     : OUT std_logic_vector(3 DOWNTO 0);

      -- chipscope debug ports
      cs_io                   : INOUT std_logic_vector(1023 DOWNTO 0);   --  for debug
      dip                     : IN  std_logic
   );
END ENTITY rld_seven_seg;



ARCHITECTURE rtl OF rld_seven_seg IS

   SIGNAL count1         : std_logic_vector(26 DOWNTO 0);
   SIGNAL count2         : std_logic_vector(26 DOWNTO 0);
   SIGNAL read_div_cnt   : std_logic_vector(READ_DIV_CNT_WIDTH-1 DOWNTO 0);
   SIGNAL init_done_reg  : std_logic;
   SIGNAL pass_reg       : std_logic;
   SIGNAL error_reg      : std_logic;
   SIGNAL valid_r1       : std_logic;
   SIGNAL disp_data      : std_logic_vector(3 DOWNTO 0);


BEGIN
   cs_io <= (OTHERS => 'Z');
   seven_seg_dp_n  <= '1';

   LED(0)  <= count2(26);     -- System clock
   LED(1)  <= init_done_reg;
   LED(2)  <= pass_reg;
   LED(3)  <= error_reg;

   disp_data  <= read_div_cnt(READ_DIV_CNT_WIDTH-1 DOWNTO READ_DIV_CNT_WIDTH-4);


   -- want to produce a slow blinking lights to check system clock
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst_n = '0' ) THEN
            count2 <= (others=>'1');   -- reset
         ELSE
            count2 <= STD_LOGIC_VECTOR( UNSIGNED(count2) + 1 );   -- increment
         END IF;

         IF ( rst_n = '0' ) THEN
            read_div_cnt <= (others => '0');
         ELSE
            IF ( valid_r1 = '1' ) THEN
               read_div_cnt <= STD_LOGIC_VECTOR( UNSIGNED(read_div_cnt) + 1 );
            ELSE
               read_div_cnt <= read_div_cnt;
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- register outputs for LEDs  
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst_n = '0' ) THEN
            init_done_reg <= '0';
            pass_reg      <= '0';
            error_reg     <= '0';
            valid_r1      <= '0';
         ELSE
            init_done_reg <= init_done;
            valid_r1 <= valid;

            IF ( init_done = '1' ) THEN
               -- if not in reset state
               error_reg <= PASS_FAIL(2) AND NOT PASS_FAIL(1);
               pass_reg  <= NOT PASS_FAIL(2) AND PASS_FAIL(1);
            ELSE
               error_reg <= error_reg;
               pass_reg  <= pass_reg;
            END IF;
         END IF;
      END IF;
   END PROCESS;


   -- output 7-Seg display   
   PROCESS (clk)
   BEGIN
      IF RISING_EDGE( clk ) THEN
         IF ( rst_n = '0' ) THEN
            seven_seg_n <= (others=>'1');     -- clear the display
         ELSE
            IF ( init_done_reg = '0' ) THEN
               seven_seg_n <= "1000110";      -- display "C" for Calibration
            ELSE
               IF ( error_reg = '1' ) THEN
                  seven_seg_n <= "0111111";   -- display "-" for Error
               ELSE
--                  IF ( dip = '1' ) THEN
                     -- update the display
                     CASE disp_data IS
                        WHEN "0000"  =>  seven_seg_n <= "1000000";
                        WHEN "0001"  =>  seven_seg_n <= "1111001";
                        WHEN "0010"  =>  seven_seg_n <= "0100100";
                        WHEN "0011"  =>  seven_seg_n <= "0110000";
                        WHEN "0100"  =>  seven_seg_n <= "0011001";
                        WHEN "0101"  =>  seven_seg_n <= "0010010";
                        WHEN "0110"  =>  seven_seg_n <= "0000010";
                        WHEN "0111"  =>  seven_seg_n <= "1111000";
                        WHEN "1000"  =>  seven_seg_n <= "0000000";
                        WHEN "1001"  =>  seven_seg_n <= "0011000";
                        WHEN "1010"  =>  seven_seg_n <= "0001000";
                        WHEN "1011"  =>  seven_seg_n <= "0000011";
                        WHEN "1100"  =>  seven_seg_n <= "1000110";
                        WHEN "1101"  =>  seven_seg_n <= "0100001";
                        WHEN "1110"  =>  seven_seg_n <= "0000110";
                        WHEN "1111"  =>  seven_seg_n <= "0001110";
                        WHEN OTHERS  =>  seven_seg_n <= "1111111";  -- default
                     END CASE;
--                  ELSE   -- output error signals
--                     seven_seg_n <= '0' & NOT cs_io(44) & NOT cs_io(46) & '1' & NOT cs_io(47) & NOT cs_io(45) & '1';
--                  END IF;
               END IF;
            END IF;
         END IF;
      END IF;
   END PROCESS;


END ARCHITECTURE rtl;
